package ctalz.cnstr;

import java.io.*;
import java.util.*;
import java.util.regex.*;
import ctalz.util.*;

public class CVCInterface extends ConstraintSolver
{
	private Process prover;
	private OutputStream toProverStream;
	private InputStream fromProverStream;
	private int resetCount = 0, flushCount = 0;
	private static int maxResetCount = Config.asInt("ConstraintSolverResetInterval", Integer.MAX_VALUE);
	
	public CVCInterface()
	{
		init();
	}
	
	private void init()
	{
		Debug.println("Starting theorem prover...");
		long time = System.nanoTime();
		try {
			ProcessBuilder pb = new ProcessBuilder(Config.asString("CVCCommand", "").split(" "));
			pb.redirectErrorStream(true);
			prover = pb.start();
			toProverStream = prover.getOutputStream();
			fromProverStream = prover.getInputStream();
		} catch (IOException e) {
			ExceptionAccumulator.addFatal(null, "cannotLaunchProver", e);
		} finally {
			timeSpentInSolver += System.nanoTime() - time;
		}
		sendToProver("PUSH;");
	}
	public void terminate()
	{
		if (prover != null) {
			prover.destroy();
			Debug.println("Theorem prover terminated");
		}
	}
	
	private void sendToProver(CharSequence str)
	{
		try {
			Debug.println("Sending to prover: " + str);
			long time = System.nanoTime();
			toProverStream.write(str.toString().getBytes());
			toProverStream.write(Utilities.EOL.getBytes());
			toProverStream.flush();
			timeSpentInSolver += System.nanoTime() - time;
		} catch (IOException e) {
			ExceptionAccumulator.addFatal(null, "proverDown", e);
		}
	}
	private String receiveFromProver()
	{
		try {
			StringBuilder buffer = new StringBuilder();
			long time = System.nanoTime();
			for (;;) {
				int data = fromProverStream.read();
				if (data < 0)
					throw new EOFException("Connection was closed");
				if (data == '\n' || data == '\r')
					break;
				buffer.append((char)data);
			}
			timeSpentInSolver += System.nanoTime() - time;
			String result = buffer.toString();
			Debug.println("Received from prover: " + result);
			return result;
		} catch (IOException e) {
			ExceptionAccumulator.addFatal(null, "proverDown", e);
			return null;
		}
	}
	private void flushProverResponse()
	{
		if (flushCount++ < 10)
			return;
		flushCount = 0;
		try {
			long time = System.nanoTime();
			for (;;) {
				int size = fromProverStream.available();
				if (size <= 0)
					break;
				fromProverStream.read(new byte[size]);
			}
			timeSpentInSolver += System.nanoTime() - time;
		} catch (IOException e) {
			ExceptionAccumulator.addFatal(null, "proverDown", e);
		}
	}
	
	private void sendDeclarations(Collection<VarExpr> vars)
	{
		if (vars.isEmpty())
			return;
		StringBuilder commands = new StringBuilder();
		int count = 0;
		for (VarExpr v : vars) {
			if (count++ > 0)
				commands.append(',');
			commands.append(v.solverName());
		}
		sendToProver(commands.append(":INT;"));
	}
	
	public void reset()
	{
		if (resetCount++ < maxResetCount) {
			sendToProver("POP;");
			sendToProver("PUSH;");
		} else {
			terminate();
			init();
			resetCount = 0;
		}
	}
	public void saveEnvironment()
	{
		sendToProver("PUSH;");
	}
	public void restoreEnvironment()
	{
		sendToProver("POP;");
	}
	public void assume(Prop prop)
	{
		if (prop == ConstProp.TRUE)
			return;
		
		StringBuilder command = new StringBuilder("ASSERT ");
		prop.appendForSolver(command);
		command.append(";");
		sendDeclarations(prop.collectFreeVariables(new HashSet<VarExpr>()));
		sendToProver(command);
		flushProverResponse();
	}
	public boolean isValid(Prop prop)
	{
		if (prop == ConstProp.TRUE || prop.evaluate() == Boolean.TRUE)
			return true;
		
		StringBuilder command = new StringBuilder("QUERY ");
		prop.appendForSolver(command);
		command.append(";");
		sendToProver("PUSH;");
		sendDeclarations(prop.collectFreeVariables(new HashSet<VarExpr>()));
		sendToProver(command);
		sendToProver("POP;");
		for (;;) {
			String response = receiveFromProver();
			boolean valid = response.contains("Valid"), invalid = response.contains("Invalid");
			if (valid && !invalid)
				return true;
			if (!valid && invalid)
				return false;
			if (response.contains("***"))
				ExceptionAccumulator.addFatal(null, "unrecognizedProverResponse");
			if (response.contains("Unknown")) {
				for (int count = 0; count < 2; ) {
					if (receiveFromProver().trim().equals(""))
						count++;
				}
				return false;
			}
		}
	}
	public Expr simplify(Expr expr)
	{
		if (expr instanceof ConstExpr || expr instanceof VarExpr)
			return expr;
		long value = expr.evaluate();
		if (value != Expr.CANNOT_EVALUATE)
			return ConstExpr.valueOf((int)value);
		
		int originalSize = expr.nodeSize();
		Collection<VarExpr> vars = expr.collectFreeVariables(new HashSet<VarExpr>());
		Map<String, VarExpr> varMap = new HashMap<String, VarExpr>();
		for (VarExpr v : vars)
			varMap.put(v.solverName(), v);
		
		StringBuilder command = new StringBuilder("TRANSFORM ");
		Pattern pat = Pattern.compile("[CV> ]*(.+)");
		expr.appendForSolver(command);
		command.append(";");
		sendToProver("PUSH;");
		sendDeclarations(vars);
		sendToProver(command);
		sendToProver("POP;");
		for (;;) {
			String response = receiveFromProver();
			Matcher matcher = pat.matcher(response);
			if (matcher.matches()) {
				String e = matcher.group(1).trim();
				if (!e.equals("")) {
					Expr s = CVCParser.parse(e, varMap);
					return (s != null && s.nodeSize() < originalSize ? s : expr);
				}
			}
			if (response.contains("***"))
				ExceptionAccumulator.addFatal(null, "unrecognizedProverResponse");
		}
	}
	public Prop simplify(Prop prop)
	{
		Boolean value = prop.evaluate();
		if (value == Boolean.TRUE)
			return ConstProp.TRUE;
		else if (value == Boolean.FALSE)
			return ConstProp.FALSE;
		return prop;
	}
}
