package ctalz.cnstr;

import java.util.*;
import ctalz.util.*;

public class OpExpr extends Expr
{
	protected Op op;
	protected Expr left, right;
	
	public static enum Op
	{
		ADD("+", 2, 2, true), SUB("-", 2, 2, true), MUL("*", 4, 2, true), DIV("/", 4, 2, true), 
		NEG("-", 6, 1, true);
		
		private String name;
		private int arity, priority;
		private boolean assocLeft;
		Op(String n, int p, int a, boolean l) { name = n; priority = p; arity = a; assocLeft = l; }
	}
	
	protected OpExpr(Op o, Expr l)
	{
		assert o.arity == 1;
		op = o; left = l;
	}
	protected OpExpr(Op o, Expr l, Expr r)
	{
		assert o.arity == 2;
		op = o; left = l; right = r;
	}
	
	public Expr left()
	{
		return left;
	}
	public Expr right()
	{
		return right;
	}
	
	public static Expr calculate(Op o, Expr l)
	{
		if (o == Op.NEG && l instanceof ConstExpr)
			return ConstExpr.valueOf(-((ConstExpr)l).value());
		return new OpExpr(o, l);
	}
	public static Expr calculate(Op o, Expr l, Expr r)
	{
		if ((o == Op.ADD || o == Op.SUB) && l instanceof ConstExpr && r instanceof ConstExpr) {
			if (o == Op.ADD)
				return ConstExpr.valueOf(((ConstExpr)l).value() + ((ConstExpr)r).value());
			else
				return ConstExpr.valueOf(((ConstExpr)l).value() - ((ConstExpr)r).value());
		}
		return new OpExpr(o, l, r);
	}
	
	protected Collection<VarExpr> collectFreeVariables(Collection<VarExpr> set)
	{
		left.collectFreeVariables(set);
		if (right != null)
			right.collectFreeVariables(set);
		return set;
	}
	
	public Expr substitute(Map<VarExpr, Expr> map)
	{
		Expr newLeft = left.substitute(map);
		if (right != null) {
			Expr newRight = right.substitute(map);
			return (newLeft == left && newRight == right ? this : new OpExpr(op, newLeft, newRight));
		} else
			return (newLeft == left ? this : new OpExpr(op, newLeft));
	}
	
	protected void appendForUser(StringBuilder out, int outerPriority, boolean longFormat)
	{
		if (right != null) {
			if (op.priority < outerPriority)
				out.append("(");
			left.appendForUser(out, op.priority | (op.assocLeft ? 0 : 1), longFormat);
			out.append(' ').append(op.name).append(' ');
			right.appendForUser(out, op.priority | (op.assocLeft ? 1 : 0), longFormat);
			if (op.priority < outerPriority)
				out.append(")");
		} else {
			if (outerPriority > 0)
				out.append('(');
			out.append(op.name);
			left.appendForUser(out, op.priority, longFormat);
			if (outerPriority > 0)
				out.append(')');
		}
	}
	protected void appendForSolver(StringBuilder out)
	{
		if (right != null) {
			out.append('(');
			left.appendForSolver(out); out.append(op.name); right.appendForSolver(out);
			out.append(')');
		} else {
			out.append('(');
			out.append(op.name); left.appendForSolver(out);
			out.append(')');
		}
	}
	
	public int nodeSize()
	{
		return (1 + left.nodeSize() + (right != null ? right.nodeSize() : 0));
	}
	public long evaluate()
	{
		long lv = left.evaluate();
		if (lv == CANNOT_EVALUATE)
			return CANNOT_EVALUATE;
		if (op == Op.NEG)
			lv = -lv;
		else {
			long rv = right.evaluate();
			if (rv == CANNOT_EVALUATE)
				return CANNOT_EVALUATE;
			switch (op) {
			case ADD: lv += rv; break;
			case SUB: lv -= rv; break;
			case MUL: lv *= rv; break;
			case DIV: if (rv == 0) return CANNOT_EVALUATE; lv /= rv; break;
			default: return CANNOT_EVALUATE;
			}
		}
		return ((lv & ~0xFFFFFFFFL) == 0 ? lv : CANNOT_EVALUATE);
	}
	public boolean deepEquals(Expr expr)
	{
		if (!(expr instanceof OpExpr))
			return false;
		OpExpr e = (OpExpr)expr;
		if (op != e.op || (right == null) != (e.right == null))
			return false;
		return (left.deepEquals(e.left) && (right == null || right.deepEquals(e.right)));
	}
}
