package ctalz.cnstr;

import java.util.*;
import ctalz.util.*;

public class RelProp extends Prop
{
	private Op op;
	private Expr left, right;
	
	public static enum Op
	{
		EQ("=", "="), NE("/=", "!="), 
		LT("<", "<"), LE("<=", "<="), GT(">", ">"), GE(">=", ">="), 
		DV("=", "|"), ND("/=", "!|");
		
		private String solverName, userName;
		Op(String p, String u) { solverName = p; userName = u; }
		public String toString() { return userName; }
	}
	
	private RelProp()
	{
	}
	public RelProp(Op o, Expr l, Expr r)
	{
		this(null, o, l, r);
	}
	public RelProp(Location loc, Op o, Expr l, Expr r)
	{
		CTALException e = initInstance(this, loc, o, l, r);
		if (e != null) {
			op = Op.EQ; left = ConstExpr.valueOf(0); right = ConstExpr.valueOf(0);
			ExceptionAccumulator.add(e);
		}
	}
	public static RelProp newOrNull(Op o, Expr l, Expr r)
	{
		RelProp result = new RelProp();
		if (initInstance(result, Location.NULL, o, l, r) != null)
			return null;
		return result;
	}
	private static CTALException initInstance(RelProp result, Location loc, Op o, Expr l, Expr r)
	{
		result.op = o; result.left = l; result.right = r;
		if (result.op == Op.DV || result.op == Op.ND) {
			if (!(result.left instanceof ConstExpr))
				return new CTALException(loc, "leftsideMul");
			while (result.right instanceof OpExpr) {
				int lv = ((ConstExpr)result.left).value();
				OpExpr rv = (OpExpr)result.right;
				if (lv != 0 && (rv.op == OpExpr.Op.ADD || rv.op == OpExpr.Op.SUB)) {
					if (rv.left instanceof ConstExpr && ((ConstExpr)rv.left).value() % lv == 0) {
						result.right = rv.right;
						continue;
					} else if (rv.right instanceof ConstExpr && ((ConstExpr)rv.right).value() % lv == 0) {
						result.right = rv.left;
						continue;
					}
				}
				break;
			}
			if (!(result.right instanceof ConstExpr) && !(result.right instanceof VarExpr))
				return new CTALException(loc, "rightsideMul");
		}
		return null;
	}
	
	public Op op()
	{
		return op;
	}
	public Expr left()
	{
		return left;
	}
	public Expr right()
	{
		return right;
	}
	
	protected Collection<VarExpr> collectFreeVariables(Collection<VarExpr> set)
	{
		left.collectFreeVariables(set);
		right.collectFreeVariables(set);
		if (op == Op.DV || op == Op.ND) {
			if (right instanceof VarExpr)
				set.add(((VarExpr)right).multipleFor(((ConstExpr)left).value()));
		}
		return set;
	}
	
	public Prop substitute(Map<VarExpr, Expr> map)
	{
		Expr newLeft = left.substitute(map), newRight = right.substitute(map);
		return (newLeft == left && newRight == right ? this : new RelProp(op, newLeft, newRight));
	}
	
	protected void appendForUser(StringBuilder out, int outerOp)
	{
		left.appendForUser(out, 0, true);
		out.append(' ').append(op.userName).append(' ');
		right.appendForUser(out, 0, true);
	}
	protected void appendForSolver(StringBuilder out)
	{
		if (op == Op.DV || op == Op.ND) {
			int div = ((ConstExpr)left).value();
			if (div == 1 || div == -1)
				out.append("(TRUE)");
			else if (right instanceof ConstExpr) {
				if (((ConstExpr)right).value() % div == 0)
					out.append("(TRUE)");
				else
					out.append("(FALSE)");
			} else {
				out.append('(');
				right.appendForSolver(out); out.append(op.solverName);
				((VarExpr)right).multipleFor(div).appendForSolver(out);
				out.append(')');
			}
		} else {
			out.append('(');
			left.appendForSolver(out); out.append(op.solverName); right.appendForSolver(out);
			out.append(')');
		}
	}
	
	public Boolean evaluate()
	{
		long lv = left.evaluate();
		if (lv == Expr.CANNOT_EVALUATE)
			return null;
		long rv = right.evaluate();
		if (rv == Expr.CANNOT_EVALUATE)
			return null;
		switch (op) {
		case EQ: return (lv == rv);
		case NE: return (lv != rv);
		case LT: return (lv < rv);
		case LE: return (lv <= rv);
		case GT: return (lv > rv);
		case GE: return (lv >= rv);
		case DV: return (rv % lv == 0);
		case ND: return (rv % lv == 0);
		}
		return null;
	}
	public boolean deepEquals(Prop prop)
	{
		if (!(prop instanceof RelProp))
			return false;
		RelProp p = (RelProp)prop;
		return (op == p.op && left.deepEquals(p.left) && right.deepEquals(p.right));
	}
}
