package ctalz.comp;

import java.io.*;
import java.util.*;
import ctalz.comp.ast.*;
import ctalz.comp.act.*;
import ctalz.mac.*;
import ctalz.util.*;

public class Main
{
	private static String uniqueHash;
	private static String encoding;
	
	public static void main(String[] args)
	{
		int exitCode = 0;
		try {
			Utilities.loadResources("ctalz/comp/LocalizedStrings", 
									"ctalz/type/LocalizedStrings", 
									"ctalz/cnstr/LocalizedStrings", 
									"ctalz/util/LocalizedStrings");
			if (args.length < 1 || args.length > 3) {
				System.err.println(Utilities.localizedString("usage"));
				System.exit(1);
			}
			String inputFile = args[0], 
				   outputFile = (args.length >= 2 ? args[1] : Utilities.changeFileSuffix(inputFile, ".c", ".cts"));
			uniqueHash = (Config.asInt("ScrambleStaticNames", 0) == 0 ? 
						  "" : "_" + Utilities.uniqueHash(inputFile + outputFile));
			encoding = (args.length >= 3 ? args[2] : null);
			
			Debug.println("Parsing...");
			List<ExtDef> defs = new Parser(inputFile, encoding).parse();
			if (ExceptionAccumulator.numberOfExceptions() > 0)
				throw new AbortError();
			
			Debug.println("Scanning...");
			constructEnv(defs);
			if (ExceptionAccumulator.numberOfExceptions() > 0)
				throw new AbortError();
			
			Debug.println("Transforming...");
			transform(defs);
			if (ExceptionAccumulator.numberOfExceptions() > 0)
				throw new AbortError();
			
			Debug.println("Compiling...");
			Program prog = compile(defs);
			if (ExceptionAccumulator.numberOfExceptions() > 0)
				throw new AbortError();
			
			emitAssembly(prog, outputFile);
		} catch (AbortError e) {
			exitCode = failureExitCode();
		}
		
		if (!ExceptionAccumulator.isEmpty())
			ExceptionAccumulator.printlnTo(System.err);
		
		System.exit(exitCode);
	}
	
	private static void constructEnv(List<ExtDef> defs)
	{
		Env env = new Env();
		ListIterator<ExtDef> it = defs.listIterator();
		FunDef.predeclare(env);
		while (it.hasNext())
			it.set(it.next().constructEnv(env));
		if (ExceptionAccumulator.numberOfExceptions() > 0)
			throw new AbortError();
		List<ExtDef> generics = new LinkedList<ExtDef>();
		for (ExtDef d : defs) {
			if (d instanceof FunDef) {
				if (((FunDef)d).needGeneric())
					generics.add(((FunDef)d).createGeneric().constructEnv(env));
			}
		}
		defs.addAll(generics);
	}
	private static void transform(List<ExtDef> defs)
	{
		List<VarDef> prepend = new LinkedList<VarDef>();
		ListIterator<ExtDef> it = defs.listIterator();
		while (it.hasNext())
			it.set(it.next().transform(prepend));
		defs.addAll(0, prepend);
		
		int print = Config.asInt("PrintTransformedCode", 0);
		if (print > 0) {
			PrintStream out = (print == 1 ? System.out : System.err);
			out.println("Transformed Code:");
			for (ExtDef d : defs) {
				out.println(d);
				out.println();
			}
		}
	}
	public static Program compile(List<ExtDef> defs)
	{
		Program prog = new Program();
		Function init = new Function(FunDef.initName, false, new LinkedList<Var>(), 
									 new Var(ActType.freshInt()), new Var(ActType.freshPackedBP()));
		prog.add(init);
		new CallExpr(Location.NULL, "_.init", new ConstExpr(Location.NULL, 0)).compile(prog, null, init);
		for (ExtDef d : defs)
			d.compile(prog, init, init);
		init.compile(prog);
		return prog;
	}
	
	public static void emitPreamble(PrintStream out) throws IOException
	{
		try {
			String printf = Program.stubName("printf"), malloc = Arch.encodeLabelName(FunDef.mallocName), alignEsp;
			if (Arch.os() == Arch.OSType.MAC)
				alignEsp = ".unsafe_native \"\\tandl $~15, %esp\", [], []" + Utilities.EOL + "\t";
			else
				alignEsp = "";
			for (String line : Utilities.loadResource("/ctalz/comp/preamble")) {
				out.println(line.replace("%printf", printf).replace("%malloc", malloc)
								.replace("%alignesp;", alignEsp).replace("%hash", uniqueHash));
			}
		} catch (FileNotFoundException e) {
			ExceptionAccumulator.addFatal(Location.NULL, "missingResource", e);
		}
	}
	private static void emitAssembly(Program prog, String outputFile)
	{
		try {
			PrintStream out;
			if (outputFile.equals("-"))
				out = System.out;
			else
				out = new PrintStream(outputFile);
			emitPreamble(out);
			prog.emitAssembly(out);
			out.flush();
			if (out.checkError())
				throw new IOException("I/O Error");
			if (!outputFile.equals("-"))
				out.close();
		} catch (IOException e) {
			ExceptionAccumulator.addFatal(Location.NULL, "ioError", e);
		}
	}
	
	public static String uniqueHash()
	{
		return uniqueHash;
	}
	public static String encoding()
	{
		return encoding;
	}
	
	public static int failureExitCode()
	{
		return 1;
	}
}
