package ctalz.comp.act;

import java.util.*;
import ctalz.cnstr.*;
import ctalz.util.*;

public abstract class ActNode
{
	final int id;
	private static int nextID;
	List<ActNode> succs, preds;
	VarSet liveIn, liveOut;
	static final int NO_EAX = 1, NO_EBX = 2, NO_ECX = 4, NO_EDX = 8, NO_EDI = 8, NO_CONST = 256;
	
	public ActNode()
	{
		id = ++nextID;
	}
	
	public final boolean isUnconditionalJump()
	{
		return (this instanceof Jump && ((Jump)this).op == Jump.Op.ALWAYS);
	}
	
	void addSuccs(ActNode follow)
	{
		succs.add(follow);
	}
	VarSet uses(boolean add, VarSet result)
	{
		return result;
	}
	void registerVars(VarSet set)
	{
	}
	VarSet defines(boolean add, VarSet result)
	{
		return result;
	}
	boolean hasSideEffects()
	{
		return false;
	}
	
	boolean foldConstant(Substitution<ActNode> map, ActNode next)
	{
		return true;
	}
	
	void coercionConvert(Program prog, List<ActNode> nodes, StackLayout stack, RegFile regs)
	{
		nodes.add(this);
	}
	void prologueConvert(Program prog, List<ActNode> nodes, StackLayout stack)
	{
		nodes.add(this);
	}
	
	static void spill(List<ActNode> nodes, StackLayout stack, RegFile regs, Operand op)
	{
		if (!(op instanceof Reg))
			return;
		Reg reg = (Reg)op;
		if (!regs.modified[reg.index] || !(regs.contents[reg.index] instanceof Var))
			return;
		Var var = (Var)regs.contents[reg.index];
		if (var.isGlobal) {
			coerce(nodes, stack, regs, reg, var.type.kind);
			nodes.add(new Asm("st<4,0>", reg, var));
		} else {
			int s = stack.find(var);
			coerce(nodes, stack, regs, reg, stack.types.get(s).kind);
			stack.types.set(s, regs.types[reg.index]);
			nodes.add(new StackAccess(StackAccess.Op.ST, stack, var, reg));
		}
		regs.modified[reg.index] = false;
	}
	static void spillAll(List<ActNode> nodes, StackLayout stack, RegFile regs)
	{
		for (Reg r : Reg.instances) {
			if (r != Reg.scratch)
				spill(nodes, stack, regs, r);
		}
	}
	
	static Operand restore(List<ActNode> nodes, StackLayout stack, RegFile regs, ActNode var, int flags)
	{
		Reg r = regs.find(var);
		if (r != null && (r.noFlags & flags) != 0) {
			spill(nodes, stack, regs, r);
			regs.set(r, null, null);
			regs.modified[r.index] = false;
			r = null;
		}
		if ((flags & NO_CONST) == 0 && var instanceof Const)
			return (Const)var;
		if (r == null) {
			r = reserve(nodes, stack, regs, flags);
			restore(nodes, stack, regs, var, r);
		} else if (var instanceof Const) {
			if (regs.types[r.index].kind != ActType.Kind.INT)
				restore(nodes, stack, regs, var, r);
			regs.types[r.index] = ActType.newInt(((Const)var).val);
		}
		regs.markUsed(r);
		return r;
	}
	static void restore(List<ActNode> nodes, StackLayout stack, RegFile regs, ActNode var, Reg reg)
	{
		if (var instanceof Const) {
			nodes.add(new Arith(Arith.Op.ID, false, reg, var, null));
			regs.types[reg.index] = ActType.newInt(((Const)var).val);
		} else if (((Var)var).isGlobal) {
			nodes.add(new Asm("ld<4,0>", reg, var));
			if (((Var)var).type.kind == ActType.Kind.INT) {
				regs.types[reg.index] = ActType.freshInt();
				coerce(nodes, regs, reg);
			} else
				regs.types[reg.index] = ((Var)var).type;
		} else {
			int s = stack.find((Var)var);
			nodes.add(new StackAccess(StackAccess.Op.LD, stack, (Var)var, reg));
			regs.types[reg.index] = stack.types.get(s);
		}
		regs.modified[reg.index] = false; regs.contents[reg.index] = var;
	}
	static Reg reserve(List<ActNode> nodes, StackLayout stack, RegFile regs, int flags)
	{
		Reg r = regs.findEmpty(flags);
		if (r == null) {
			r = regs.leastRecentlyUsed(flags);
			spill(nodes, stack, regs, r);
		}
		regs.set(r, null, null);
		regs.modified[r.index] = false;
		regs.markUsed(r);
		return r;
	}
	
	static ActType coerce(List<ActNode> nodes, StackLayout stack, RegFile regs, Operand op, ActType.Kind to)
	{
		if (!(op instanceof Reg)) {
			assert to == ActType.Kind.INT;
			return ActType.newInt(((Const)op).val);
		}
		Reg reg = (Reg)op;
		ActType from = regs.types[reg.index];
		Var v;
		switch (to) {
		case INT:
			if (from.kind == ActType.Kind.INT)
				return regs.setWithoutModifying(reg, from);
			break;
		case BP:
			switch (from.kind) {
			case INT:
				if (from.val != ConstExpr.valueOf(0))
					break;
				from = packNullBP(nodes, stack, regs, reg);
				/* fall thru */
			case PACKED_BP:
				return regs.setWithoutModifying(reg, unpackBP(nodes, stack, regs, reg, from));
			case BP:
				return regs.setWithoutModifying(reg, from);
			}
			break;
		case PACKED_BP:
			switch (from.kind) {
			case INT:
				if (from.val != ConstExpr.valueOf(0))
					break;
				return regs.setWithoutModifying(reg, packNullBP(nodes, stack, regs, reg));
			case BP:
				return regs.setWithoutModifying(reg, packBP(nodes, stack, regs, reg, from));
			case PACKED_BP:
				return regs.setWithoutModifying(reg, from);
			}
			break;
		case VAR:
			if (from.kind == ActType.Kind.VAR)
				return regs.setWithoutModifying(reg, from);
			break;
		}
		assert false : "Invalid coercion from " + from + " to " + to;
		return null;
	}
	static void coerce(List<ActNode> nodes, RegFile regs, Reg reg)
	{
		assert regs.types[reg.index].kind == ActType.Kind.INT;
		nodes.add(new Asm("let", reg, regs.types[reg.index].val));
	}
	private static Var freshVar()
	{
		return new Var(null, "v" + Utilities.uniqueID());
	}
	
	private static ActType packBP(List<ActNode> nodes, StackLayout stack, RegFile regs, Reg reg, ActType unpacked)
	{
		nodes.add(new Asm("packh", reg, "ExBaseRec", "[" + unpacked.len + "," + unpacked.meta + "]"));
		nodes.add(new Asm("rollh", reg, "BaseRec"));
		nodes.add(new Asm("guardh", reg, "NullBaseRec[" + unpacked.val + "]"));
		nodes.add(new Asm("packr", reg, "BasePtr", "[" + unpacked.val + "]"));
		return ActType.newPackedBP(unpacked.len, unpacked.meta);
	}
	private static ActType packNullBP(List<ActNode> nodes, StackLayout stack, RegFile regs, Reg reg)
	{
		nodes.add(new Asm("packr", reg, "BasePtr", "[0]"));
		return ActType.freshPackedBP();
	}
	private static ActType unpackBP(List<ActNode> nodes, StackLayout stack, RegFile regs, Reg reg, ActType packed)
	{
		Expr val = new VarExpr("v" + Utilities.uniqueID());
		nodes.add(new Asm("unpackr", reg, "[" + val + "]"));
		nodes.add(new Jump(Jump.Op.EQ, false, Jump.abortLabel, new Const(0), reg).coercionConvert(stack, regs));
		nodes.add(new Asm("unguardfh", reg));
		nodes.add(new Asm("unrollh", reg));
		nodes.add(new Asm("unpackh", reg, "[" + packed.len.toShortString() + "," + packed.meta.toShortString() + "]"));
		return ActType.newBP(val, packed.len, packed.meta);
	}
	
	static void loadLenMetaToScratch(List<ActNode> nodes, Reg bp, boolean loadLen)
	{
		nodes.add(new Asm((loadLen ? "ld<4,0>" : "ld<4,1>"), Reg.scratch, (loadLen ? "0(" : "4(") + bp + ")"));
	}
	
	String toAsmString()
	{
		assert false : this;
		return "";
	}
	static String toAsmString(String... tokens)
	{
		StringBuilder result = new StringBuilder("\t");
		int count = 0;
		for (String t : tokens) {
			if (t == null)
				continue;
			if (count > 1)  result.append(',');
			if (count >= 1) result.append(' ');
			result.append(t);
			count++;
		}
		return result.toString();
	}
}
