package ctalz.comp.act;

import java.util.*;
import ctalz.cnstr.*;
import ctalz.util.*;

public class Arith extends ActNode
{
	public final Op op;
	public final boolean isUnsigned;
	ActNode dest, lhs, rhs;
	
	public enum Op
	{
		ID("mov"), 
		MINUS("neg", OpExpr.Op.NEG), ADD("add", OpExpr.Op.ADD), SUB("sub", OpExpr.Op.SUB), 
		MUL("mul", "umul", NO_CONST), DIV("div", "udiv", NO_CONST), MOD("mod", "umod", NO_CONST), 
		AND("and"), OR("or"), XOR("xor"), LSHIFT("lsl"), RSHIFT("asr", "lsr"), 
		EXTEND1("se1", "ze1", NO_EDI), EXTEND2("se2", "ze2"), 
		NOT("not");
		
		final String signedMnemonic, unsignedMnemonic;
		final OpExpr.Op exprOp;
		final int noFlags;
		Op(String m) { this(m, m, null, 0); }
		Op(String m, OpExpr.Op o) { this(m, m, o, 0); }
		Op(String s, String u) { this(s, u, null, 0); }
		Op(String s, String u, int f) { this(s, u, null, f); }
		Op(String s, String u, OpExpr.Op o, int f) {
			signedMnemonic = s; unsignedMnemonic = u; exprOp = o; noFlags = f;
		}
	}
	
	public Arith(Op o, boolean u, ActNode d, ActNode l, ActNode r)
	{
		assert (l instanceof Var) || (l instanceof Const) || (l instanceof Reg);
		assert r == null || (r instanceof Var) || (r instanceof Const) || (r instanceof Reg);
		op = o; isUnsigned = u; dest = (d instanceof Reg ? d : ((VarRef)d).var); lhs = l; rhs = r;
	}
	
	void registerVars(VarSet set)
	{
		set.register(dest); set.register(lhs); set.register(rhs);
	}
	VarSet uses(boolean add, VarSet result)
	{
		if (lhs instanceof Var) { if (add) result.add((Var)lhs); else result.remove((Var)lhs); }
		if (rhs instanceof Var) { if (add) result.add((Var)rhs); else result.remove((Var)rhs); }
		return result;
	}
	VarSet defines(boolean add, VarSet result)
	{
		if (dest instanceof Var) { if (add) result.add((Var)dest); else result.remove((Var)dest); }
		return result;
	}
	boolean hasSideEffects()
	{
		return (dest instanceof Var && ((Var)dest).isVolatile);
	}
	
	boolean foldConstant(Substitution<ActNode> map, ActNode next)
	{
		dest = Manipulator.swapDest(dest, next);
		lhs = map.get(lhs); rhs = map.get(rhs);
		map.removeKeyValue(dest);
		
		boolean done = false;
		if (op == Op.ID)
			map.put(dest, lhs);
		else if (op == Op.MUL || op == Op.DIV) {
			if (isConst(lhs, 0) || (op == Op.MUL && isConst(rhs, 0))) { map.put(dest, new Const(0)); done = true; }
			else if (op == Op.MUL && isConst(lhs, 1)) { map.put(dest, rhs); done = true; }
			else if (isConst(rhs, 1)) { map.put(dest, lhs); done = true; }
		} else if (op == Op.ADD || op == Op.SUB) {
			if (op == Op.ADD && isConst(lhs, 0)) { map.put(dest, rhs); done = true; }
			else if (isConst(rhs, 0)) { map.put(dest, lhs); done = true; }
		}
		if (!done && lhs instanceof Const && rhs instanceof Const) {
			long lv = ((Const)lhs).extendedLongValue(!isUnsigned), rv = ((Const)rhs).extendedLongValue(!isUnsigned);
			try {
				switch (op) {
				case ADD: map.put(dest, new Const((int)(lv + rv))); break;
				case SUB: map.put(dest, new Const((int)(lv - rv))); break;
				case MUL: map.put(dest, new Const((int)(lv * rv))); break;
				case DIV: map.put(dest, new Const((int)(lv / rv))); break;
				case MOD: map.put(dest, new Const((int)(lv % rv))); break;
				case AND: map.put(dest, new Const((int)(lv & rv))); break;
				case OR:  map.put(dest, new Const((int)(lv | rv))); break;
				case XOR: map.put(dest, new Const((int)(lv ^ rv))); break;
				case LSHIFT: map.put(dest, new Const((int)(lv << rv))); break;
				case RSHIFT: map.put(dest, new Const((int)(lv >> rv))); break;
				}
			} catch (Exception e) {
			}
		} else if (!done && lhs instanceof Const && rhs == null) {
			long lv = ((Const)lhs).extendedLongValue(!isUnsigned);
			switch (op) {
			case MINUS: map.put(dest, new Const((int)(-lv))); break;
			case EXTEND1:
				map.put(dest, new Const((lv & 0x80) != 0 ? ~0xFF | ((int)lv & 0xFF) : (int)lv & 0xFF)); break;
			case EXTEND2:
				map.put(dest, new Const((lv & 0x8000) != 0 ? ~0xFFFF | ((int)lv & 0xFFFF) : (int)lv & 0xFFFF)); break;
			case NOT: map.put(dest, new Const((int)(~lv))); break;
			}
		}
		return (op != Op.ID || dest != lhs);
	}
	private static boolean isConst(ActNode node, int val)
	{
		return (node instanceof Const && ((Const)node).val == val);
	}
	
	void coercionConvert(Program prog, List<ActNode> nodes, StackLayout stack, RegFile regs)
	{
		Reg dreg = (Reg)restore(nodes, stack, regs, lhs, NO_CONST | op.noFlags);
		Operand rreg;
		spill(nodes, stack, regs, dreg);
		rreg = (rhs != null ? restore(nodes, stack, regs, rhs, op.noFlags) : null);
		
		if (op != Op.ID) {
			ActType lt = coerce(nodes, stack, regs, dreg, ActType.Kind.INT), rt = null;
			if (rhs != null)
				rt = coerce(nodes, stack, regs, rreg, ActType.Kind.INT);
			if (op.exprOp != null) {
				regs.set(dreg, dest, ActType.newInt(rt != null ? 
						 OpExpr.calculate(op.exprOp, lt.val, rt.val) : OpExpr.calculate(op.exprOp, lt.val)));
			} else
				regs.set(dreg, dest, ActType.freshInt());
			
			dest = lhs = dreg;
			if (rhs != null)
				rhs = rreg;
			nodes.add(this);
			if (op.exprOp == null)
				coerce(nodes, regs, dreg);
		} else {
			coerce(nodes, stack, regs, dreg, ((Var)dest).type.kind);
			if (((Var)dest).type.kind != ActType.Kind.PACKED_BP)
				regs.set(dreg, dest, regs.types[dreg.index]);
			else
				regs.set(dreg, dest, ((Var)dest).type);
		}
	}
	
	public String toString()
	{
		return ("(" + (isUnsigned ? op.unsignedMnemonic : op.signedMnemonic) + " " + dest + " " + 
				lhs + (rhs != null ? " " + rhs : "") + ")");
	}
	String toAsmString()
	{
		String m = (isUnsigned ? op.unsignedMnemonic : op.signedMnemonic);
		if (op == Op.ID)
			return toAsmString(m, dest.toAsmString(), lhs.toAsmString());
		else if (op == Op.EXTEND1 || op == Op.EXTEND2)
			return toAsmString(m, dest.toAsmString(), dest.toAsmString());
		else
			return toAsmString(m, dest.toAsmString(), (rhs != null ? rhs.toAsmString() : null));
	}
}
