package ctalz.comp.act;

import java.util.*;
import ctalz.comp.*;
import ctalz.mac.*;
import ctalz.util.*;

public class Call extends ActNode
{
	protected Var userDest, baseDest;
	private ActNode addr, base;
	private String label;
	private boolean isLabelGlobal;
	private List<ActNode> args;
	private List<Var> calleeArgs;
	private StackLayout savedStack;
	private RegFile savedRegs;
	private List<ActType> pushedTypes;
	private int stackAlign = -1;
	
	public Call(Var u, Var b, String l, boolean g, List<ActNode> a, List<Var> c)
	{
		userDest = u; baseDest = b; label = l; isLabelGlobal = g; args = a; calleeArgs = c;
	}
	public Call(Var u, Var b, ActNode f, ActNode g, List<ActNode> a, List<Var> c)
	{
		userDest = u; baseDest = b; addr = f; base = g; args = a; calleeArgs = c;
	}
	
	void registerVars(VarSet set)
	{
		set.register(userDest); set.register(baseDest); set.register(addr); set.register(base);
		for (ActNode a : args)
			set.register(a);
	}
	VarSet uses(boolean add, VarSet result)
	{
		if (addr instanceof Var) { if (add) result.add((Var)addr); else result.remove((Var)addr); }
		if (base instanceof Var) { if (add) result.add((Var)base); else result.remove((Var)base); }
		for (ActNode a : args)
			if (a instanceof Var) { if (add) result.add((Var)a); else result.remove((Var)a); }
		return result;
	}
	VarSet defines(boolean add, VarSet result)
	{
		if (userDest != null) { if (add) result.add(userDest); else result.remove(userDest); }
		if (baseDest != null) { if (add) result.add(baseDest); else result.remove(baseDest); }
		return result;
	}
	boolean hasSideEffects()
	{
		return true;
	}
	
	boolean foldConstant(Substitution<ActNode> map, ActNode next)
	{
		if (userDest != null) map.removeKeyValue(userDest);
		if (baseDest != null) map.removeKeyValue(baseDest);
		userDest = (Var)Manipulator.swapDest(userDest, next);
		baseDest = (Var)Manipulator.swapDest(baseDest, next);
		addr = map.get(addr); base = map.get(base);
		for (int i = 0; i < args.size(); i++)
			args.set(i, map.get(args.get(i)));
		return true;
	}
	
	void coercionConvert(Program prog, List<ActNode> nodes, StackLayout stack, RegFile regs)
	{
		spillAll(nodes, stack, regs);
		nodes.add(this);
		pushedTypes = new ArrayList<ActType>();
		for (int i = args.size() - 1; i >= 0; i--)
			pushedTypes.add(push(nodes, stack, regs, args.get(i), calleeArgs.get(i).type, true));
		if (label == null) {
			Reg areg = (Reg)restore(nodes, stack, regs, addr, NO_CONST), 
				breg = (Reg)restore(nodes, stack, regs, base, NO_CONST);
			coerce(nodes, stack, regs, areg, ActType.Kind.INT);
			coerce(nodes, stack, regs, breg, ActType.Kind.BP);
			loadLenMetaToScratch(nodes, breg, true);
			boundsCheck(nodes, stack, regs, areg, breg, Reg.scratch);
			loadLenMetaToScratch(nodes, breg, false);
			LoadUB.useMeta(nodes, false, Reg.scratch);
			label = Reg.scratch.toAsmString();
		} else
			label = Arch.encodeLabelName(label) + (isLabelGlobal ? "" : Main.uniqueHash());
		nodes.add(this);
		savedStack = stack.saveTypes();
		savedRegs = regs.copy();
		stack.extraElems = 0; stack.extraSize = 0;
		
		regs.clear();
		if (userDest != null)
			regs.set(Reg.userReturn, userDest, userDest.type);
		if (baseDest != null)
			regs.set(Reg.baseReturn, baseDest, baseDest.type);
	}
	static void boundsCheck(List<ActNode> nodes, StackLayout stack, RegFile regs, Reg user, Reg base, Reg len)
	{
		nodes.add(new Jump(Jump.Op.NE, false, Jump.abortLabel, user, base).coercionConvert(stack, regs));
		nodes.add(new Jump(Jump.Op.LE, false, Jump.abortLabel, new Const(0), len).coercionConvert(stack, regs));
	}
	static ActType push(List<ActNode> nodes, StackLayout stack, RegFile regs, ActNode var, ActType type, 
						boolean doCoerce)
	{
		Operand r = restore(nodes, stack, regs, var, (!doCoerce || type.kind == ActType.Kind.INT ? 0 : NO_CONST));
		int s = type.sizeof();
		ActType t = (doCoerce ? coerce(nodes, stack, regs, r, type.kind) : type);
		nodes.add(new Asm("push<" + s + ">", r));
		stack.extraElems++;
		stack.extraSize += s;
		return t;
	}
	
	void prologueConvert(Program prog, List<ActNode> nodes, StackLayout stack)
	{
		if (stackAlign < 0)
			stackAlign = alignStack(nodes, savedStack.completeMissingTypes());
		else {
			Label cont = new Label("cont");
			calleeArgs = new ArrayList<Var>(calleeArgs);
			Collections.reverse(calleeArgs);
			savedStack.completeMissingTypes();
			savedStack.types.addAll(pushedTypes);
			
			Pair<LabelType, LabelType> destType = Function.entryLabelType(calleeArgs, userDest, baseDest);
			String funApp, contApp;
			List<Pair<String, String>> rename;
			savedStack.types.add(ActType.newRetAddr(destType.snd));
			funApp = destType.fst.calculateApplyParams(nodes, savedStack, savedRegs);
			savedStack.types.remove(savedStack.types.size() - 1);
			rename = cont.prologueConvertForContinuation(savedStack, savedRegs, userDest, baseDest);
			contApp = cont.labelType().calculateApplyParamsForContinuation(nodes, savedStack, savedRegs);
			
			nodes.add(new Asm("apply", "addrof(" + label + ")", funApp));
			nodes.add(new Asm("apply", "addrof(" + cont.toAsmString() + ")", contApp));
			nodes.add(new Asm("call", label));
			nodes.add(cont);
			for (Pair<String, String> p : rename)
				nodes.add(new Asm("let", p.fst, p.snd));
			freeStack(nodes, savedStack, stackAlign);
		}
	}
	static int alignStack(List<ActNode> nodes, StackLayout stack)
	{
		int align = Arch.stackPointerAlignment(), size = stack.completeMissingTypes().allocatedSize() + 4;
		if (size % align != 0) {
			int ins = stack.vars.size(), add = ((size / align + 1) * align - size) / 4;
			StringBuilder ss = new StringBuilder(), is = new StringBuilder();
			for (int i = 0; i < add; i++) {
				Var v = new Var(ActType.freshInt());
				if (i != 0) { ss.append(','); is.append(','); }
				ss.append('4'); is.append(v.type.val); stack.vars.add(ins, v);
			}
			nodes.add(new Asm("salloc<" + ss + ">", "$" + (add * 4), "[" + is + "]"));
			stack.completeMissingTypes();
			return add;
		}
		return 0;
	}
	static void freeStack(List<ActNode> nodes, StackLayout stack, int aligned)
	{
		for (int i = 0; i < aligned; i++) {
			stack.vars.remove(stack.vars.size() - 1);
			stack.types.remove(stack.vars.size() - 1);
		}
		if (stack.extraElems + aligned > 0)
			nodes.add(new Asm("sfree<" + (stack.extraElems + aligned) + ">", "$" + (stack.extraSize + aligned * 4)));
	}
	
	public String toString()
	{
		return ("(call" + (isLabelGlobal ? "/g " : " ") + (userDest == null ? "*" : userDest) + " " + 
				(baseDest == null ? "*" : baseDest) + " " + (label != null ? "\"" + label + "\"" : addr + " " + base) + 
				(args.isEmpty() ? "" : " " + Utilities.enumerate(args, " ")) + ")");
	}
}
