package ctalz.comp.act;

import java.util.*;
import ctalz.comp.*;
import ctalz.util.*;

public class Jump extends ActNode
{
	final Op op;
	boolean isUnsigned;
	Label dest;
	ActNode lhs, rhs;
	boolean isFallThrough;
	private StackLayout savedStack;
	private RegFile savedRegs;
	public static final Label abortLabel = new Label(null), misalignAbortLabel = new Label(null);
	
	public enum Op
	{
		ALWAYS("jmp"), 
		EQ("beq"), NE("bne"), LT("blt", "bltu"), LE("ble", "bleu"), GT("bgt", "bgtu"), GE("bge", "bgeu"), 
		DV("bdv"), ND("bnd");
		
		final String signedMnemonic, unsignedMnemonic;
		Op(String m) { signedMnemonic = unsignedMnemonic = m; }
		Op(String s, String u) { signedMnemonic = s; unsignedMnemonic = u; }
	}
	
	public Jump(ActNode d)
	{
		this(Op.ALWAYS, false, d, null, null);
	}
	public Jump(Op o, boolean u, ActNode d, ActNode l, ActNode r)
	{
		assert l == null || (l instanceof Var) || (l instanceof Const) || (l instanceof Reg);
		assert r == null || (r instanceof Var) || (r instanceof Const) || (r instanceof Reg);
		op = o; isUnsigned = u; dest = (Label)d; lhs = l; rhs = r;
	}
	
	void addSuccs(ActNode follow)
	{
		if (op != Op.ALWAYS)
			succs.add(follow);
		if (dest != abortLabel && dest != misalignAbortLabel)
			succs.add(dest);
	}
	void registerVars(VarSet set)
	{
		set.register(dest); set.register(lhs); set.register(rhs);
	}
	VarSet uses(boolean add, VarSet result)
	{
		if (lhs instanceof Var) { if (add) result.add((Var)lhs); else result.remove((Var)lhs); }
		if (rhs instanceof Var) { if (add) result.add((Var)rhs); else result.remove((Var)rhs); }
		return result;
	}
	boolean hasSideEffects()
	{
		return true;
	}
	
	boolean foldConstant(Substitution<ActNode> map, ActNode next)
	{
		lhs = map.get(lhs); rhs = map.get(rhs);
		return true;
	}
	
	Boolean isConstantlyTaken()
	{
		if (!(lhs instanceof Const) || !(rhs instanceof Const))
			return null;
		long lv = ((Const)lhs).extendedLongValue(!isUnsigned), rv = ((Const)rhs).extendedLongValue(!isUnsigned);
		switch (op) {
		case EQ: return (lv == rv);
		case NE: return (lv != rv);
		case LT: return (lv < rv);
		case LE: return (lv <= rv);
		case GT: return (lv > rv);
		case GE: return (lv >= rv);
		}
		return null;
	}
	
	void coercionConvert(Program prog, List<ActNode> nodes, StackLayout stack, RegFile regs)
	{
		spillAll(nodes, stack, regs);
		Operand lreg = (lhs != null && !(lhs instanceof Reg) ? restore(nodes, stack, regs, lhs, 0) : null), 
				rreg = (rhs != null && !(lhs instanceof Reg) ? restore(nodes, stack, regs, rhs, NO_CONST) : null);
		if (lreg != null)
			coerce(nodes, stack, regs, lreg, ActType.Kind.INT);
		if (rreg != null)
			coerce(nodes, stack, regs, rreg, ActType.Kind.INT);
		
		lhs = lreg; rhs = rreg;
		savedStack = stack.saveTypes();
		savedRegs = regs.copy();
		nodes.add(this);
	}
	Jump coercionConvert(StackLayout stack, RegFile regs)
	{
		savedStack = stack.saveTypes();
		savedRegs = regs.copy();
		return this;
	}
	void prologueConvert(Program prog, List<ActNode> nodes, StackLayout stack)
	{
		if (abortLabel.name == null) {
			abortLabel.name = "_$abort" + Main.uniqueHash();
			misalignAbortLabel.name = "_$abort_misalign" + Main.uniqueHash();
		}
		savedStack.completeMissingTypes();
		if (dest != abortLabel && dest != misalignAbortLabel) {
			String app = ((Label)dest).labelType().calculateApplyParams(nodes, savedStack, savedRegs);
			nodes.add(new Asm("apply", "addrof(" + dest.toAsmString() + ")", app));
		} else
			nodes.add(new Asm("apply", "addrof(" + dest.name + ")", "*", "*"));
		nodes.add(this);
		if (op != Op.ALWAYS)
			nodes.add(new Asm("recover", (dest != abortLabel && dest != misalignAbortLabel ? dest : dest.name)));
	}
	
	public String toString()
	{
		return ("(" + (isUnsigned ? op.unsignedMnemonic : op.signedMnemonic) + " " + 
				(dest != null ? dest : "<abort>") + 
				(lhs != null ? " " + lhs : "") + (rhs != null ? " " + rhs : "") + ")");
	}
	String toAsmString()
	{
		if (isFallThrough)
			return toAsmString("fallthru");
		return toAsmString((isUnsigned ? op.unsignedMnemonic : op.signedMnemonic), 
						   (lhs != null ? lhs.toAsmString() : null), 
						   (rhs != null ? rhs.toAsmString() : null), 
						   (dest != abortLabel && dest != misalignAbortLabel ? dest.toAsmString() : dest.name));
	}
}
