package ctalz.comp.act;

import java.util.*;
import ctalz.cnstr.*;
import ctalz.util.*;

class LabelType
{
	private List<String> generalVars, generalIndices;
	private List<ActType> stack;
	private ActType[] regs;
	
	public LabelType(StackLayout s, RegFile r)
	{
		this(s.saveTypes().resetTypes(), r, true);
	}
	private LabelType(StackLayout s, RegFile r, boolean generalizeSp)
	{
		generalVars = new ArrayList<String>();
		generalIndices = new ArrayList<String>();
		stack = new ArrayList<ActType>();
		regs = new ActType[Reg.instances.length];
		stack.add(ActType.newVar(s.tail));
		if (generalizeSp)
			generalVars.add(stack.get(0).val.toShortString());
		for (int i = 0; i < s.types.size(); i++)
			stack.add(register(s.types.get(i), true));
		for (int i = 0; i < regs.length; i++)
			regs[i] = r.types[i] = register(r.types[i], (generalizeSp || !Reg.isCalleeSave(i)));
	}
	private ActType register(ActType t, boolean generalize)
	{
		if (t == null) {
			ActType v = ActType.newVar("v" + Utilities.uniqueID());
			if (generalize)
				generalVars.add(v.val.toShortString());
			return v;
		}
		if (hasIndex(t) && generalize) {
			if (t.val instanceof VarExpr)
				generalIndices.add(t.val.toShortString());
			else {
				for (VarExpr v : t.val.freeVariables())
					generalIndices.add(v.toShortString());
			}
		}
		return t;
	}
	
	static LabelType forReturnAddress(StackLayout stack, Var userRet, Var baseRet)
	{
		RegFile regs = new RegFile(stack = stack.saveTypes().resetTypes());
		if (userRet != null)
			regs.types[Reg.userReturn.index] = userRet.type;
		if (baseRet != null)
			regs.types[Reg.baseReturn.index] = baseRet.type;
		return new LabelType(stack, regs, false);
	}
	static LabelType forContinuation(StackLayout stack, Var userRet, Var baseRet)
	{
		RegFile regs = new RegFile();
		if (userRet != null)
			regs.types[Reg.userReturn.index] = userRet.type;
		if (baseRet != null)
			regs.types[Reg.baseReturn.index] = baseRet.type;
		regs.types[Reg.scratch.index] = ActType.freshInt();
		return new LabelType(stack, regs, true);
	}
	
	private String calculateApplyParams(List<ActNode> nodes, StackLayout s, RegFile r, boolean forCont)
	{
		Map<String, String> map = new HashMap<String, String>();
		for (int i = 1, offset = 0; ; i++) {
			ActType to = stack.get(stack.size() - i);
			if (i == stack.size()) {
				List<ActType> sp = new ArrayList<ActType>(s.types.subList(0, s.types.size() - i + 1));
				sp.add(0, ActType.newVar(s.tail));
				map.put(to.toAsmString(), "sptr{" + reverseEnumerate(sp) + "}");
				break;
			}
			if (!forCont || i > s.extraElems) {
				ActType from = s.types.get(s.types.size() - i);
				from = coerceStack(nodes, i - 1, offset, from, to.kind);
				if (hasIndex(to))
					unify(map, to.val, from.val);
				offset += from.sizeof();
			}
		}
		for (int i = 0; i < regs.length; i++) {
			if (forCont && !Reg.isCalleeSave(i))
				continue;
			ActType to = regs[i], from = r.types[i];
			if (to.kind == ActType.Kind.VAR || i == Reg.scratch.index)
				map.put(to.val.toShortString(), Reg.instances[i].toString());
			else {
				from = coerceReg(nodes, r, i, from, to.kind);
				if (hasIndex(to))
					unify(map, to.val, from.val);
			}
		}
		
		List<String> vars = new ArrayList<String>(generalVars.size()), 
					 inds = new ArrayList<String>(generalIndices.size());
		for (String v : generalVars) {
			v = map.get(v);
			vars.add(v != null ? v : "*");
		}
		for (String v : generalIndices) {
			v = map.get(v);
			inds.add(v != null ? v : "*");
		}
		return "[" + Utilities.enumerate(vars, ",") + "],[" + Utilities.enumerate(inds, ",") + "]";
	}
	String calculateApplyParams(List<ActNode> nodes, StackLayout s, RegFile r)
	{
		return calculateApplyParams(nodes, s, r, false);
	}
	String calculateApplyParamsForContinuation(List<ActNode> nodes, StackLayout s, RegFile r)
	{
		return calculateApplyParams(nodes, s, r, true);
	}
	private static void unify(Map<String, String> map, Expr to, Expr from)
	{
		if (to instanceof VarExpr)
			map.put(to.toShortString(), from.toShortString());
		else if (to instanceof OpExpr) {
			OpExpr t = (OpExpr)to, f = (OpExpr)from;
			unify(map, t.left(), f.left());
			if (t.right() != null)
				unify(map, t.right(), f.right());
		}
	}
	
	private static ActType coerceStack(List<ActNode> nodes, int elem, int offset, ActType from, ActType.Kind to)
	{
		assert from.kind == to : "Invalid stack coercion from " + from.kind + " to " + to.name;
		return from;
	}
	private static ActType coerceReg(List<ActNode> nodes, RegFile regs, int index, ActType from, ActType.Kind to)
	{
		if ((from == null || from.kind != to) && to == ActType.Kind.INT) {
			regs.types[index] = ActType.freshInt();
			nodes.add(new Asm("let", Reg.instances[index], regs.types[index].val));
			return regs.types[index];
		}
		assert from.kind == to : "Invalid register coercion from " + from.kind + " to " + to.name;
		return from;
	}
	
	private boolean hasIndex(ActType t)
	{
		return (t.kind == ActType.Kind.INT && !(t.val instanceof ConstExpr));
	}
	
	public String toString()
	{
		StringBuilder result = new StringBuilder("code");
		result.append(generalVars()).append(generalIndices());
		result.append('[');
		for (int i = 0; i < regs.length; i++)
			result.append(Reg.instances[i]).append('=').append(regs[i].toAsmString()).append(',');
		result.append(Reg.sp.name).append("=sptr{").append(reverseEnumerate(stack)).append("}][][]");
		return result.toString();
	}
	public String generalVars()
	{
		return "[" + Utilities.enumerate(generalVars, ",") + "]";
	}
	public String generalIndices()
	{
		return "[" + Utilities.enumerate(generalIndices, ",") + "]";
	}
	private static String reverseEnumerate(List<ActType> types)
	{
		StringBuilder result = new StringBuilder();
		int count = 0;
		for (int i = types.size() - 1; i >= 0; i--) {
			if (count++ > 0)
				result.append(',');
			result.append(types.get(i).toAsmString());
		}
		return result.toString();
	}
}
