package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.comp.ctype.*;
import ctalz.util.*;

public class AssignExpr extends Expr
{
	private Op op;
	private Expr lhs, rhs;
	
	public AssignExpr(Op o, Expr l, Expr r)
	{
		super(l.loc);
		op = o; lhs = l; rhs = r;
	}
	public AssignExpr(Location loc, Op o, Expr l)
	{
		super(loc);
		op = o; lhs = l;
	}
	
	enum Op
	{
		SIMPLE(100, " = ", null, true), 
		ADD(100, " += ", Arith.Op.ADD, true), 
		SUB(100, " -= ", Arith.Op.SUB, true), 
		MUL(100, " *= ", Arith.Op.MUL, true), 
		DIV(100, " /= ", Arith.Op.DIV, true), 
		MOD(100, " %= ", Arith.Op.MOD, true), 
		LSHIFT(100, " <<= ", Arith.Op.LSHIFT, true), 
		RSHIFT(100, " >>= ", Arith.Op.RSHIFT, true), 
		BIT_OR(100, " |= ", Arith.Op.OR, true), 
		BIT_XOR(100, " ^= ", Arith.Op.XOR, true), 
		BIT_AND(100, " &= ", Arith.Op.AND, true), 
		PRE_INCR(800, "++", null, true), 
		PRE_DECR(800, "--", null, true), 
		POST_INCR(900, "++", null, true), 
		POST_DECR(900, "--", null, true), 
		POST_ADD(100, " <+=> ", Arith.Op.ADD, false), 
		POST_SUB(100, " <-=> ", Arith.Op.SUB, false);
		
		final int priority;
		final String name;
		final Arith.Op arithOp;
		final boolean pre;
		Op(int p, String n, Arith.Op o, boolean r) { priority = p; name = n; arithOp = o; pre = r; }
	}
	
	int priority()
	{
		return op.priority;
	}
	
	Expr constructEnv(Context ctx, boolean pointerTaken, Env env)
	{
		if (op == Op.PRE_INCR)
			return new AssignExpr(Op.ADD, lhs, new ConstExpr(loc, 1)).constructEnv(ctx, false, env);
		if (op == Op.PRE_DECR)
			return new AssignExpr(Op.SUB, lhs, new ConstExpr(loc, 1)).constructEnv(ctx, false, env);
		if (op == Op.POST_INCR)
			return new AssignExpr(Op.POST_ADD, lhs, new ConstExpr(loc, 1)).constructEnv(ctx, false, env);
		if (op == Op.POST_DECR)
			return new AssignExpr(Op.POST_SUB, lhs, new ConstExpr(loc, 1)).constructEnv(ctx, false, env);
		
		boolean canPtr = (op == Op.SIMPLE || op == Op.ADD || op == Op.SUB || op == Op.POST_ADD || op == Op.POST_SUB);
		lhs = lhs.constructEnv(Context.LVALUE, false, env);
		CType lt = lhs.type;
		if (lt.isConst())
			ExceptionAccumulator.add(lhs.loc, "constAssign", lt);
		else if (lt instanceof FunType || lt instanceof ArrayType || lt == BasicType.VOID)
			ExceptionAccumulator.add(lhs.loc, "lvalueType", lt);
		else if (lt.sizeof() <= 0)
			ExceptionAccumulator.add(lhs.loc, "sizeNotKnown", lt);
		else if (!canPtr && lt.isPtr())
			ExceptionAccumulator.add(lhs.loc, "operandType", op.name.trim(), lhs.type);
		else if (!lt.isNumber() && !lt.isPtr())
			reportUnsupportedFeature("aggrAssign");
		rhs = rhs.constructEnv(env);
		if (op != Op.SIMPLE && lt.isPtr()) {
			rhs = UnaryExpr.integralPromote(rhs, op.name.trim());
			if (!rhs.type.isIntegral())
				ExceptionAccumulator.add(rhs.loc, "operandType", op.name.trim(), rhs.type);
			else {
				int s = ((PtrType)lhs.type).base.sizeof();
				if (s <= 0)
					ExceptionAccumulator.add(loc, "sizeNotKnown", ((PtrType)lhs.type).base);
				rhs = new BinaryExpr(BinaryExpr.Op.MUL, rhs, new ConstExpr(rhs.loc, s)).typeConvert(rhs.type);
			}
		} else if (!rhs.type.equals(lhs.type))
			rhs = new CastExpr(rhs.loc, lhs.type, rhs);
		return typeConvert(ctx, lhs.type);
	}
	
	public Expr transform(List<VarDef> prepend)
	{
		lhs = lhs.transform(prepend);
		rhs = rhs.transform(prepend);
		return this;
	}
	
	public ConstValue eval()
	{
		return null;
	}
	
	ActValue compileAsRvalue(Program prog, Function fun)
	{
		ActValue l = lhs.compileAsLvalue(prog, fun), r = rhs.compileAsRvalue(prog, fun);
		ActValue lr = (op == Op.SIMPLE ? null : lvalueToRvalue(fun, lhs.type, l));
		ActNode u, b;
		Var pu = null, pb = null;
		if (!op.pre) {
			fun.add(new Arith(Arith.Op.ID, false, new VarRef(pu = new Var(type)), lr.userValue, null));
			if (lr.baseValue != null) {
				fun.add(new Arith(Arith.Op.ID, false, 
								  new VarRef(pb = new Var(ActType.freshPackedBP())), lr.baseValue, null));
			}
		}
		if (op == Op.SIMPLE) {
			u = r.userValue; b = r.baseValue;
		} else {
			u = new Var(type);
			fun.add(new Arith(op.arithOp, lhs.type.isUnsigned(), new VarRef((Var)u), lr.userValue, r.userValue));
			if (lr.baseValue != null) {
				b = new Var(ActType.freshPackedBP());
				fun.add(new Arith(Arith.Op.ID, false, new VarRef((Var)b), lr.baseValue, null));
			} else
				b = null;
		}
		if (l.userValue instanceof VarRef) {
			fun.add(new Arith(Arith.Op.ID, false, l.userValue, u, null));
			if (l.baseValue != null)
				fun.add(new Arith(Arith.Op.ID, false, l.baseValue, b, null));
		} else
			fun.add(new StoreUB(lhs.type.sizeof(), u, b, l.userValue, l.baseValue));
		return (op.pre ? new ActValue(u, b) : new ActValue(pu, pb));
	}
	
	public String toString()
	{
		if (rhs != null) {
			String l = lhs.toString(), r = rhs.toString();
			if (!(lhs instanceof AssignExpr) && !(lhs instanceof UnaryExpr) && 
				!(lhs instanceof IndexExpr) && !(lhs instanceof IdentExpr))
				l = "(" + l + ")";
			if (rhs.priority() < priority())
				r = "(" + r + ")";
			return l + op.name + r;
		} else {
			String l = lhs.toString();
			if (lhs.priority() <= priority())
				l = "(" + l + ")";
			return (op == Op.PRE_INCR || op == Op.PRE_DECR ? op.name + l : l + op.name);
		}
	}
}
