package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.comp.ctype.*;
import ctalz.util.*;

public class BinaryExpr extends Expr
{
	private Op op;
	private Expr lhs, rhs;
	
	public BinaryExpr(Op o, Expr l, Expr r)
	{
		super(l.loc);
		op = o; lhs = l; rhs = r;
	}
	
	enum Op
	{
		COMMA(0, ", ", null), 
		ADD(602, " + ", Arith.Op.ADD), 
		SUB(603, " - ", Arith.Op.SUB), 
		MUL(604, " * ", Arith.Op.MUL), 
		DIV(605, " / ", Arith.Op.DIV);
		
		final int priority;
		final String name;
		final Arith.Op arithOp;
		Op(int p, String n, Arith.Op o) { priority = p; name = n; arithOp = o; }
	}
	
	int priority()
	{
		return op.priority;
	}
	
	Expr constructEnv(Context ctx, boolean pointerTaken, Env env)
	{
		lhs = lhs.constructEnv(env); rhs = rhs.constructEnv(env);
		CType lt = lhs.type, rt = rhs.type;
		if (op == Op.COMMA)
			return typeConvert(ctx, rt);
		if (op == Op.ADD) {
			if (lt.isPtr())
				return new PtrArithExpr(PtrArithExpr.Op.PTR_PLUS_INT, lhs, rhs).constructEnv(ctx, false, env);
			else if (rt.isPtr())
				return new PtrArithExpr(PtrArithExpr.Op.PTR_PLUS_INT, lhs, rhs).constructEnv(ctx, false, env);
		}
		if (op == Op.SUB) {
			if (lt.isPtr() && rt.isPtr())
				return new PtrArithExpr(PtrArithExpr.Op.PTR_MINUS_PTR, lhs, rhs).constructEnv(ctx, false, env);
			else if (lt.isPtr())
				return new PtrArithExpr(PtrArithExpr.Op.PTR_MINUS_INT, lhs, rhs).constructEnv(ctx, false, env);
		}
		Pair<Expr, Expr> p = arithmeticConvert(lhs, rhs, op.name.trim());
		lhs = p.fst; rhs = p.snd;
		return typeConvert(ctx, lhs.type);
	}
	
	public Expr transform(List<VarDef> prepend)
	{
		lhs = lhs.transform(prepend);
		rhs = rhs.transform(prepend);
		return this;
	}
	
	public ConstValue eval()
	{
		if (op == Op.COMMA)
			return null;
		
		ConstValue l = lhs.eval(), r = rhs.eval();
		if (l == null || r == null)
			return null;
		switch (op) {
		case ADD: return l.add(r);
		case SUB: return l.sub(r);
		case MUL: return l.mul(r);
		case DIV: return l.div(r);
		}
		return null;
	}
	
	ActValue compileAsRvalue(Program prog, Function fun)
	{
		return compileAsRvalue(prog, fun, false);
	}
	ActValue compileAsRvalue(Program prog, Function fun, boolean keepBase)
	{
		if (op == Op.COMMA) {
			lhs.compileAsRvalue(prog, fun);
			return rhs.compileAsRvalue(prog, fun);
		}
		Var u = null;
		ActNode b = null;
		ActValue l = lhs.compileAsRvalue(prog, fun), r = rhs.compileAsRvalue(prog, fun);
		fun.add(new Arith(op.arithOp, lhs.type.isUnsigned(), new VarRef(u = new Var(type)), l.userValue, r.userValue));
		if (l.baseValue != null) {
			if (keepBase) {
				b = new Var(ActType.freshPackedBP());
				fun.add(new Arith(Arith.Op.ID, false, new VarRef((Var)b), l.baseValue, null));
			} else
				b = Const.NULL_BP;
		} else
			assert false;
		return new ActValue(u, b);
	}
	
	static Pair<Expr, Expr> arithmeticConvert(Expr lhs, Expr rhs, String op)
	{
		if (!lhs.type.isNumber())
			ExceptionAccumulator.add(lhs.loc, "operandType", op, lhs.type);
		if (!rhs.type.isNumber())
			ExceptionAccumulator.add(rhs.loc, "operandType", op, rhs.type);
		CType x = lhs.type.dequalify(), y = rhs.type.dequalify(), result;
		if (x == BasicType.DOUBLE || y == BasicType.DOUBLE)
			result = BasicType.DOUBLE;
		else if (x == BasicType.FLOAT || y == BasicType.FLOAT)
			result = BasicType.FLOAT;
		else if (x == BasicType.ULONG || y == BasicType.ULONG)
			result = BasicType.ULONG;
		else if ((x == BasicType.LONG && y == BasicType.UINT) || (x == BasicType.UINT && y == BasicType.LONG))
			result = BasicType.ULONG;
		else if (x == BasicType.LONG || y == BasicType.LONG)
			result = BasicType.LONG;
		else if (x == BasicType.UINT || y == BasicType.UINT)
			result = BasicType.UINT;
		else
			result = BasicType.INT;
		return new Pair<Expr, Expr>((x == result ? lhs : new CastExpr(lhs.loc, result, lhs)), 
									(y == result ? rhs : new CastExpr(rhs.loc, result, rhs)));
	}
	
	public String toString()
	{
		String l = lhs.toString(), r = rhs.toString();
		if (lhs.priority() < priority())
			l = "(" + l + ")";
		if (rhs.priority() <= priority())
			r = "(" + r + ")";
		return l + op.name + r;
	}
}
