package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.comp.ctype.*;
import ctalz.util.*;

public class BinaryIntExpr extends Expr
{
	private Op op;
	private Expr lhs, rhs;
	
	public BinaryIntExpr(Op o, Expr l, Expr r)
	{
		super(l.loc);
		op = o; lhs = l; rhs = r;
	}
	
	enum Op
	{
		BIT_OR(400, " | ", Arith.Op.OR), 
		BIT_XOR(401, " ^ ", Arith.Op.XOR), 
		BIT_AND(402, " & ", Arith.Op.AND), 
		LSHIFT(600, " << ", Arith.Op.LSHIFT), 
		RSHIFT(601, " >> ", Arith.Op.RSHIFT), 
		MOD(606, " % ", Arith.Op.MOD);
		
		final int priority;
		final String name;
		final Arith.Op arithOp;
		Op(int p, String n, Arith.Op o) { priority = p; name = n; arithOp = o; }
	}
	
	int priority()
	{
		return op.priority;
	}
	
	Expr constructEnv(Context ctx, boolean pointerTaken, Env env)
	{
		lhs = lhs.constructEnv(env); rhs = rhs.constructEnv(env);
		if (op == Op.LSHIFT || op == Op.RSHIFT) {
			lhs = UnaryExpr.integralPromote(lhs, op.name.trim());
			rhs = UnaryExpr.integralPromote(rhs, op.name.trim());
		} else {
			Pair<Expr, Expr> p = BinaryExpr.arithmeticConvert(lhs, rhs, op.name.trim());
			lhs = p.fst; rhs = p.snd;
		}
		if (!lhs.type.isIntegral())
			ExceptionAccumulator.add(lhs.loc, "operandType", op.name.trim(), lhs.type);
		if (!rhs.type.isIntegral())
			ExceptionAccumulator.add(rhs.loc, "operandType", op.name.trim(), rhs.type);
		return typeConvert(ctx, lhs.type);
	}
	
	public Expr transform(List<VarDef> prepend)
	{
		lhs = lhs.transform(prepend);
		rhs = rhs.transform(prepend);
		return this;
	}
	
	public ConstValue eval()
	{
		ConstValue l = lhs.eval(), r = rhs.eval();
		if (l == null || r == null)
			return null;
		try {
			switch (op) {
			case BIT_OR: return l.bitOr(r);
			case BIT_XOR: return l.bitXor(r);
			case BIT_AND: return l.bitAnd(r);
			case LSHIFT: return l.shiftLeft(r);
			case RSHIFT: return l.shiftRight(r);
			case MOD: return l.mod(r);
			}
			return l;
		} catch (Exception e) {
			return null;
		}
	}
	
	ActValue compileAsRvalue(Program prog, Function fun)
	{
		Var u = new Var(type);
		ActValue l = lhs.compileAsRvalue(prog, fun), r = rhs.compileAsRvalue(prog, fun);
		fun.add(new Arith(op.arithOp, lhs.type.isUnsigned(), new VarRef(u), l.userValue, r.userValue));
		return new ActValue(u, Const.NULL_BP);
	}
	
	public String toString()
	{
		String l = lhs.toString(), r = rhs.toString();
		if (lhs.priority() < priority())
			l = "(" + l + ")";
		if (rhs.priority() <= priority())
			r = "(" + r + ")";
		return l + op.name + r;
	}
}
