package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.comp.ctype.*;
import ctalz.util.*;

public class CallExpr extends Expr
{
	private Expr fun;
	private String genericFunName;
	private List<Expr> args;
	private boolean transformVarArg;
	
	public CallExpr(Expr f, List<Expr> a)
	{
		this(f, a, true);
	}
	CallExpr(Expr f, List<Expr> a, boolean t)
	{
		super(f.loc);
		fun = f; args = a; transformVarArg = t;
	}
	public CallExpr(Location l, String n, Expr a)
	{
		super(l);
		genericFunName = n; args = Utilities.singletonArrayList(a);
		type = new PtrType(BasicType.VOID);
	}
	
	int priority()
	{
		return 900;
	}
	
	public Expr constructEnv(Context ctx, boolean pointerTaken, Env env)
	{
		if (fun instanceof IdentExpr) {
			String n = ((IdentExpr)fun).name;
			Entity f = env.getEntityOrNull(n);
			if (f == null)
				reportUnsupportedFeature("implicitFunDecl");
		}
		ListIterator<Expr> it = args.listIterator();
		while (it.hasNext())
			it.set(it.next().constructEnv(env));
		if (fun != null) {
			if (fun instanceof IdentExpr)
				fun = ((IdentExpr)fun).constructEnv(Context.NORMAL, false, env, true);
			else
				fun = fun.constructEnv(env);
			if (!fun.type.isPtr() || !(((PtrType)fun.type).base instanceof FunType)) {
				ExceptionAccumulator.add(fun.loc, "nonFunCall", fun.type);
				return typeConvert(ctx, BasicType.INT);
			} else {
				castArgs((FunType)((PtrType)fun.type).base);
				return typeConvert(ctx, ((FunType)((PtrType)fun.type).base).ret);
			}
		}
		return this;
	}
	private void castArgs(FunType f)
	{
		if (args.size() < f.args.size()) {
			ExceptionAccumulator.add(loc, "tooFewArgs");
			return;
		}
		ListIterator<Expr> at = args.listIterator();
		Iterator<Pair<String, CType>> ft = f.args.iterator();
		while (at.hasNext()) {
			Expr a = at.next();
			if (!ft.hasNext()) {
				if (f.isVarArgs) {
					at.set(a = a.type.isNumber() ? UnaryExpr.integralPromote(a, "()") : a);
					if (a.type.dequalify() instanceof StructType)
						reportUnsupportedFeature(a.loc, "aggrArg");
				} else {
					ExceptionAccumulator.add(loc, "tooManyArgs");
					return;
				}
			} else
				at.set(a = new CastExpr(a.loc, ft.next().snd, a));
			if (a.type.sizeof() < 0)
				ExceptionAccumulator.add(a.loc, "sizeNotKnown", a.type);
		}
	}
	
	public Expr transform(List<VarDef> prepend)
	{
		if (fun != null)
			fun = fun.transform(prepend);
		ListIterator<Expr> it = args.listIterator();
		while (it.hasNext())
			it.set(it.next().transform(prepend));
		FunDef fdef = funDef();
		if (fdef != null && fdef.isNative)
			return this;
		transformVarArg(prepend);
		return transformGeneric(prepend);
	}
	private void transformVarArg(List<VarDef> prepend)
	{
		if (!transformVarArg || fun == null)
			return;
		FunType ft = (FunType)((PtrType)fun.type).base;
		if (!ft.isVarArgs || ft.args.isEmpty())
			return;
		
		int fixedArgs = ft.args.size();
		int size = 0;
		for (int i = fixedArgs; i < args.size(); i++)
			size += args.get(i).type.sizeof();
		
		Entity store = new Entity(Entity.Kind.VAR, "$vararg" + Utilities.uniqueID(), new PtrType(BasicType.CHAR));
		Expr storeVar = new IdentExpr(loc, store.name, store, store.type), result = storeVar;
		VarDef storeDef = VarDef.newMallocDef(loc, store.name, size, false, false, store);
		prepend.add(storeDef);
		for (int i = args.size() - 1; i >= fixedArgs; i--) {
			size -= args.get(i).type.sizeof();
			result = concatArgStore(args.get(i), storeVar, size, result);
		}
		args = args.subList(0, fixedArgs);
		args.add(result);
	}
	private Expr transformGeneric(List<VarDef> prepend)
	{
		if (fun == null || funDef() != null)
			return this;
		
		int size = 0;
		for (Expr a : args)
			size += Math.max(4, a.type.sizeof());
		
		Entity store = new Entity(Entity.Kind.VAR, "$args" + Utilities.uniqueID(), new PtrType(BasicType.CHAR));
		Expr storeVar = new IdentExpr(loc, store.name, store, store.type);
		VarDef storeDef = VarDef.newMallocDef(loc, store.name, size, false, false, store);
		prepend.add(storeDef);
		
		Expr result = new CallExpr(fun, Collections.singletonList(storeVar)).setType(new PtrType(type));
		if (type.dequalify() != BasicType.VOID)
			result = new UnaryExpr(loc, UnaryExpr.Op.DEREF, result).setType(type);
		for (int i = args.size() - 1; i >= 0; i--) {
			size -= Math.max(4, args.get(i).type.sizeof());
			result = concatArgStore(args.get(i), storeVar, size, result);
		}
		return result;
	}
	private static Expr concatArgStore(Expr src, Expr dest, int offset, Expr rest)
	{
		return new BinaryExpr(BinaryExpr.Op.COMMA, 
				   new AssignExpr(AssignExpr.Op.SIMPLE, 
					   new UnaryExpr(src.loc, UnaryExpr.Op.DEREF, 
						   new CastExpr(src.loc, new PtrType(src.type), 
							   new PtrArithExpr(PtrArithExpr.Op.PTR_PLUS_INT, 
								   dest, new ConstExpr(src.loc, offset)))).setType(src.type), 
					   src), rest).setType(rest.type);
	}
	
	public ConstValue eval()
	{
		return null;
	}
	
	ActValue compileAsRvalue(Program prog, Function f)
	{
		FunDef fdef = funDef();
		boolean isNative = (fdef != null && fdef.isNative);
		List<ActNode> argVars = new ArrayList<ActNode>();
		for (Expr a : args) {
			ActValue v = a.compileAsRvalue(prog, f);
			argVars.add(v.userValue); if (!isNative && v.baseValue != null) argVars.add(v.baseValue);
		}
		if (fun == null) {
			Var u = new Var(ActType.freshInt()), b = new Var(ActType.freshPackedBP());
			f.add(new Call(u, b, genericFunName, false, argVars, FunDef.genericFunArgVars));
			return new ActValue(u, b);
		} else if (isNative) {
			ActNode u = (fdef.retEntity.userValue != null ? new Var(ActType.freshInt()) : null), 
					b = (fdef.retEntity.baseValue != null ? Const.NULL_BP : null);
			String n = fdef.name();
			List<ActType> argTypes = new ArrayList<ActType>();
			for (Expr a : args)
				argTypes.add(ActType.fromCType(a.type));
			f.add(new NativeCall(prog, (Var)u, n, argVars, argTypes));
			return new ActValue(u, b);
		} else if (fdef != null) {
			Var u = (fdef.retEntity.userValue != null ? new Var(ActType.freshInt()) : null), 
				b = (fdef.retEntity.baseValue != null ? new Var(ActType.freshPackedBP()) : null);
			f.add(new Call(u, b, fdef.name(), fdef.isGlobal, argVars, fdef.argVars()));
			return new ActValue(u, b);
		}
		ActValue fv = fun.compileAsRvalue(prog, f);
		Var u = new Var(type), b = (type.hasBaseValue() ? new Var(ActType.freshPackedBP()) : null);
		f.add(new Call(u, b, fv.userValue, fv.baseValue, argVars, FunDef.genericFunArgVars));
		return new ActValue(u, b);
	}
	
	private FunDef funDef()
	{
		if (!(fun instanceof IdentExpr))
			return null;
		Entity e = ((IdentExpr)fun).entity;
		if (e.kind == Entity.Kind.FUN)
			return e.fun;
		return null;
	}
	
	public String toString()
	{
		StringBuilder result = new StringBuilder();
		int count = 0;
		if (fun != null) {
			if (fun.priority() <= priority())
				result.append('(').append(fun.toString()).append(')');
			else
				result.append(fun.toString());
		} else
			result.append(genericFunName);
		result.append('(');
		for (Expr a : args) {
			if (count++ > 0)
				result.append(", ");
			if (a.priority() == 0)
				result.append('(').append(a.toString()).append(')');
			else
				result.append(a.toString());
		}
		return result.append(')').toString();
	}
}
