package ctalz.comp.ast;

import java.util.*;
import ctalz.comp.ctype.*;
import ctalz.util.*;

public class Decl
{
	public Location loc;
	public Kind kind;
	public String name;
	public Decl base;
	public List<TypeQual> quals;
	public Expr expr;
	public Params params;
	public Init init;
	
	public enum Kind
	{
		ID, PTR, ARRAY, BIT_FIELD, ENUM, INIT, FUN;
	}
	
	public Decl(Location l)
	{
		loc = l; kind = Kind.ID;
	}
	public Decl(Location l, String n)
	{
		loc = l; kind = Kind.ID; name = n;
	}
	public Decl(Location l, Kind k, List<TypeQual> q)
	{
		loc = l; kind = k; quals = q;
	}
	public Decl(Location l, Kind k, Expr e)
	{
		loc = l; kind = k; expr = e;
	}
	public Decl(Kind k, Decl d, Expr e)
	{
		loc = d.loc; kind = k; base = d; expr = e;
	}
	public Decl(Location l, Params p)
	{
		loc = l; kind = Kind.FUN; params = p;
	}
	public Decl(Location l, Decl b, Params p)
	{
		loc = l; kind = Kind.FUN; base = b; params = p;
	}
	public Decl(Location l, List<String> p)
	{
		loc = l; kind = Kind.FUN; params = new Params(p);
	}
	public Decl(Decl d, Init i)
	{
		loc = d.loc; kind = Kind.INIT; base = d; init = i;
	}
	public static Decl newPtr(Location l, Decl b)
	{
		Decl result = new Decl(l, Kind.PTR, new LinkedList<TypeQual>());
		result.base = b;
		return result;
	}
	
	public Decl basemost()
	{
		Decl d = this;
		while (d.base != null)
			d = d.base;
		return d;
	}
	public Decl setBasemost(Decl b)
	{
		Decl d = this;
		while (d.base != null)
			d = d.base;
		d.base = b;
		return this;
	}
	public Decl addDecl(Decl d)
	{
		Decl prev = null, next = this;
		while (next.base != null) {
			prev = next;
			next = next.base;
		}
		d.setBasemost(next);
		if (prev != null) {
			prev.base = d;
			return this;
		}
		return d;
	}
	
	public CType constructCType(Env env, CType b)
	{
		switch (kind) {
		case PTR: b = new PtrType(base.constructCType(env, b)); break;
		case ARRAY:
			b = base.constructCType(env, b);
			if (expr != null) {
				expr = expr.constructEnv(env);
				ConstValue v = expr.eval();
				if (v == null || !v.type.isIntegral())
					ExceptionAccumulator.add(expr.loc, "shouldBeIntConst");
				else
					b = new ArrayType(expr.loc, b, (int)v.intValue);
			} else
				b = new ArrayType(loc, b);
			break;
		case FUN: b = params.constructCType(env, base.constructCType(env, b)); break;
		case BIT_FIELD: ExceptionAccumulator.addFatal(loc, "unsupported", Utilities.localizedString("bitField"));
		default: if (base != null) b = base.constructCType(env, b); break;
		}
		if (quals != null) {
			for (TypeQual q : quals) {
				if (q.kind == TypeQual.Kind.CONST)
					b = b.constType();
			}
		}
		return b;
	}
	
	public String toString()
	{
		Decl b = basemost();
		return toString(b.name == null ? "" : b.name, false);
	}
	private String toString(String trunk, boolean trunkIsPtr)
	{
		String t = (trunk.equals("") ? "" : "(" + trunk + ")");
		switch (kind) {
		case ID: return trunk;
		case PTR:
			if (quals.isEmpty())
				return base.toString("*" + trunk, true);
			else
				return base.toString("* " + Utilities.enumerate(quals, " ") + " " + trunk, true);
		case ARRAY:
			return base.toString((trunkIsPtr ? t : trunk) + "[" + (expr != null ? expr.toString() : "") + "]", false);
		case BIT_FIELD:
			return base + " : " + expr;
		case ENUM:
			return base + " = " + expr;
		case INIT:
			return base + " = " + init;
		case FUN:
			return base.toString((trunkIsPtr ? t : trunk) + "(" + params + ")", false);
		default:
			assert false;
			return "";
		}
	}
}
