package ctalz.comp.ast;

import java.util.*;
import ctalz.comp.act.*;
import ctalz.comp.ctype.*;
import ctalz.mac.*;
import ctalz.util.*;

public class FunDef extends ExtDef
{
	private List<Spec> specs;
	private Decl decl;
	private List<VarDef> oldParams;
	private CompoundStmt body;
	final boolean isGlobal;
	public boolean isNative;
	private Entity entity;
	private List<String> argNames;
	private List<Entity> argEntities;
	Entity retEntity;
	private List<Var> argVars;
	Var genericUser, genericBase;
	public static final List<Var> genericFunArgVars;
	public static final String varArgName = "__vararg__";
	public static final String mallocName = "__malloc__", initName = "_$init";
	public static final String genericFunPrefix = ".generic_", genericPtrPrefix = ".genptr_";
	
	static {
		genericFunArgVars = new ArrayList<Var>();
		genericFunArgVars.add(new Var(ActType.freshInt()));
		genericFunArgVars.add(new Var(ActType.freshPackedBP()));
	}
	
	public FunDef(Location l, List<Spec> s, Decl d, List<VarDef> p, CompoundStmt b, boolean n)
	{
		super(l);
		specs = s; decl = d; oldParams = p; body = b; isNative = n;
		isGlobal = (isNative || ScSpec.find(specs, ScSpec.Kind.STATIC) == null);
		if (decl.kind != Decl.Kind.FUN)
			ExceptionAccumulator.add(decl.loc, "nonFunDecl");
		if (body != null)
			body.stmts.add(new LabelStmt(loc, "$end", new ExprStmt(loc)));
	}
	public String name()
	{
		return decl.basemost().name;
	}
	
	public static void predeclare(Env env)
	{
		new FunDef(Location.NULL, 
				   Arrays.asList((Spec)new ScSpec(Location.NULL, ScSpec.Kind.STATIC), 
								 (Spec)new TypeSpec(Location.NULL, TypeSpec.Kind.VOID)), 
				   genericFunDecl(Location.NULL, initName, false), null, null, false)
			.constructEnv(env);
		new FunDef(Location.NULL, 
				   Arrays.asList((Spec)new ScSpec(Location.NULL, ScSpec.Kind.STATIC), 
								 (Spec)new TypeSpec(Location.NULL, TypeSpec.Kind.VOID)), 
				   genericFunDecl(Location.NULL, mallocName, true), null, null, false)
			.constructEnv(env);
	}
	
	public FunDef constructEnv(Env env)
	{
		if (oldParams != null || decl.params.oldParams != null) {
 			reportUnsupportedFeature("oldStyleFun");
			return this;
		}
		
		int count = 0;
		FunType type = (FunType)decl.constructCType(env, Spec.constructCType(env, specs));
		String name = name();
		if (body != null)
			name = VarDef.validateName(loc, name);
		if (type.ret.isConst())
			ExceptionAccumulator.add(loc, "qualifiedRet");
		else if (type.ret.dequalify() instanceof ArrayType)
			ExceptionAccumulator.add(loc, "arrayRet");
		else if (type.ret.dequalify() instanceof StructType)
			reportUnsupportedFeature("aggrRet");
		if (!overridePrototype(loc, env, name, new Entity(this, name, type)))
			env.addEntity(loc, name, entity);
		
		env = new Env(env, true);
		argNames = new ArrayList<String>();
		argEntities = new ArrayList<Entity>();
		for (Pair<String, CType> p : ((FunType)type).args) {
			count++;
			argNames.add(p.fst);
			if (p.fst == null && body != null)
				ExceptionAccumulator.add(loc, "missingArgName", count);
			else {
				Entity e = new Entity(Entity.Kind.VAR, p.fst, p.snd);
				argEntities.add(e);
				if (p.fst != null)
					env.addEntity(loc, p.fst, e);
			}
			if (p.snd.dequalify() instanceof StructType)
				reportUnsupportedFeature("aggrArg");
		}
		if (type.isVarArgs) {
			Entity e = new Entity(Entity.Kind.VAR, varArgName, new PtrType(BasicType.CHAR));
			argEntities.add(e);
			env.addEntity(loc, varArgName, e);
		}
		
		retEntity = new Entity(Entity.Kind.VAR, "$retVal", type.ret.dequalify());
		if (retEntity.type != BasicType.VOID) {
			retEntity.userValue = new Var(ActType.fromCType(retEntity.type), retEntity.name, false, false, true);
			if (retEntity.type.hasBaseValue())
				retEntity.baseValue = new Var(ActType.freshPackedBP(), retEntity.name + ".b", false, false, true);
		}
		env.addEntity(loc, retEntity.name, retEntity);
		
		genericUser = new Var(ActType.freshInt(), genericPtrPrefix + entity.name, 
							  true, (!isGlobal && body != null), false);
		genericBase = new Var(ActType.freshPackedBP(), genericPtrPrefix + entity.name + ".b", 
							  true, (!isGlobal && body != null), false);
		
		if (body != null)
			body = body.constructEnv(false, env);
		return this;
	}
	private boolean overridePrototype(Location loc, Env env, String name, Entity e)
	{
		Entity proto = env.getEntityOrNull(name);
		if (proto != null && proto.kind == Entity.Kind.FUN && proto.fun.body == null) {
			if (!proto.type.equals(e.type) || proto.fun.isGlobal != isGlobal)
				ExceptionAccumulator.add(loc, "incompatiblePrototype");
			entity = proto;
			entity.fun = this;
			return true;
		}
		entity = e;
		return false;
	}
	
	public FunDef transform(List<VarDef> prepend)
	{
		if (body == null)
			return this;
		
		List<Stmt> p = new ArrayList<Stmt>();
		ListIterator<Entity> it = argEntities.listIterator();
		while (it.hasNext()) {
			Entity e = it.next();
			if (e.isPointerTaken) {
				Entity newEntity = new Entity(Entity.Kind.VAR, e.name, e.type);
				p.add(new DefStmt(new VarDef(loc, e, 
											 new IdentExpr(loc, newEntity.name, newEntity, newEntity.type))));
				it.set(newEntity);
			}
		}
		body.stmts.addAll(0, p);
		
		prepend = new ArrayList<VarDef>();
		body = body.transform(prepend);
		for (int i = prepend.size() - 1; i >= 0; i--)
			body.stmts.add(0, new DefStmt(prepend.get(i)));
		return this;
	}
	
	public void compile(Program prog, Function init, Function fun)
	{
		if (body == null)
			return;
		
		fun = new Function(entity.name, isGlobal, argVars(), retEntity.userValue, retEntity.baseValue);
		prog.add(fun);
		body.compile(prog, init, fun);
		if (ExceptionAccumulator.numberOfExceptions() > 0)
			throw new AbortError();
		fun.compile(prog);
		
		if (!entity.name.startsWith(genericFunPrefix) && needGeneric())
			prog.add(new Global(genericFunPrefix + entity.name, genericUser.name, genericBase.name, isGlobal));
	}
	
	List<Var> argVars()
	{
		if (argVars == null) {
			Iterator<String> it = argNames.iterator();
			argVars = new ArrayList<Var>();
			for (Entity e : argEntities) {
				if (e.kind == Entity.Kind.VAR) {
					argVars.add(e.userValue = new Var(ActType.fromCType(e.type), e.name));
					if (e.type.hasBaseValue())
						argVars.add(e.baseValue = new Var(ActType.freshPackedBP(), e.name + ".b"));
				}
			}
		}
		return argVars;
	}
	
	boolean hasGeneric()
	{
		if (isNative)
			return false;
		String n = name();
		return (!n.equals(mallocName) && !n.equals(initName));
	}
	public boolean needGeneric()
	{
		return (hasGeneric() && (isGlobal || entity.isPointerTaken));
	}
	public FunDef createGeneric()
	{
		FunType ft = (FunType)entity.type;
		List<Spec> retSpec = new LinkedList<Spec>(Spec.collectScSpecs(specs));
		retSpec.add(new TypeSpec(loc, TypeSpec.Kind.VOID));
		Decl d = genericFunDecl(loc, genericFunPrefix + entity.name, true);
		List<Stmt> stmts = new ArrayList<Stmt>();
		stmts.add(new DefStmt(new VarDef(Collections.singletonList((Spec)new TypeSpec(loc, new PtrType(ft.ret))), 
										 Collections.singletonList(new Decl(new Decl(loc, "$ret"), 
											 new Init(new CallExpr(loc, mallocName, 
														  new ConstExpr(loc, Math.max(0, ft.ret.sizeof())))))))));
		List<Pair<String, CType>> ftArgs = new ArrayList<Pair<String, CType>>(ft.args);
		List<Expr> as = new ArrayList<Expr>();
		int offset = 0;
		if (ft.isVarArgs)
			ftArgs.add(new Pair<String, CType>(varArgName, new PtrType(BasicType.CHAR)));
		for (Pair<String, CType> p : ftArgs) {
			as.add(new UnaryExpr(loc, UnaryExpr.Op.DEREF, 
								 new CastExpr(loc, new PtrType(p.snd), 
											  new BinaryExpr(BinaryExpr.Op.ADD, 
															 new IdentExpr(loc, "$arg"), new ConstExpr(loc, offset)))));
			offset += Math.max(4, p.snd.sizeof());
		}
		if (ft.ret.dequalify() != BasicType.VOID) {
			stmts.add(new ExprStmt(new AssignExpr(AssignExpr.Op.SIMPLE, 
											  new UnaryExpr(loc, UnaryExpr.Op.DEREF, new IdentExpr(loc, "$ret")), 
											  new CallExpr(new IdentExpr(loc, entity.name), as, false))));
		} else
			stmts.add(new ExprStmt(new CallExpr(new IdentExpr(loc, entity.name), as, false)));
		stmts.add(new ReturnStmt(loc, new IdentExpr(loc, "$ret")));
		return new FunDef(loc, retSpec, d, null, new CompoundStmt(loc, stmts), false);
	}
	private static Decl genericFunDecl(Location loc, String name, boolean hasArg)
	{
		Params p;
		if (hasArg) {
			p = new Params(Collections.singletonList(new SpecDecl(
					Collections.singletonList((Spec)new TypeSpec(loc, TypeSpec.Kind.CHAR)), 
					Decl.newPtr(loc, new Decl(loc, "$arg")))), false);
		} else
			p = new Params(new ArrayList<SpecDecl>(), false);
		return new Decl(loc, Decl.newPtr(loc, new Decl(loc, name)), p);
	}
	
	public String toString()
	{
		return (Utilities.enumerate(specs, " ") + (specs.isEmpty() ? "" : " ") + decl + 
				(isNative ? " = <native>;" : 
				 Utilities.EOL + 
				 (oldParams == null ? "" : Utilities.enumerate(oldParams, ";" + Utilities.EOL) + ";" + Utilities.EOL) + 
				 (body == null ? "" : body.toString(1))));
	}
}
