package ctalz.comp.ast;

import java.io.*;
import ctalz.util.*;
import static ctalz.comp.ast.Parser.*;

%%

%class Lexer
%implements Parser.yyInput
%unicode
%type Token

%{

private String fileName = "<unknown>";
private int lineNumber = 1, charNumber = 1;
private StringBuilder lineText = new StringBuilder();
private boolean isString;
private StringBuilder string;
private Location stringLoc;
private Token token;
private Parser parser;

public void setFileName(String file)
{
	fileName = file;
}
public Lexer setParser(Parser p)
{
	parser = p;
	return this;
}

public boolean advance() throws IOException
{
	token = yylex();
	if (token == null)
		return false;
	return true;
}
public int token()
{
	return token.type;
}
public Token value()
{
	return token;
}
public void close()
{
	try {
		for (int n = lineNumber; lineNumber == n && advance(); )
			;
		yyclose();
	} catch (IOException e) {
	}
}

private Location accept(String str)
{
	Location loc = loc();
	charNumber += str.length();
	lineText.append(str);
	return loc;
}
private void advanceLine()
{
	lineNumber++;
	charNumber = 1;
	lineText = new StringBuilder();
}
public Location loc()
{
	return new Location(fileName, lineNumber, charNumber, lineText);
}
public Location locWithoutChar()
{
	return new Location(fileName, lineNumber, lineText);
}

private Token constructStringToken(Location loc, String contents)
{
	if (isString)
		return new Token(loc, contents, STR_LITERAL, contents);
	else {
		int value = 0;
		for (int i = 0; i < contents.length(); i++) {
			char chr = contents.charAt(i);
			if (chr > 0xFF)
				value = (value << 16) + chr;
			else
				value = (value << 8) + chr;
		}
		return new Token(loc, contents, INT_LITERAL, value);
	}
}

%}

SP = [ \t]
EOL = \n|\r|\r\n
ZEROX = "0x" | "0X"
DIG = [0-9]
DIG16 = [0-9a-fA-F]
DOT_DIG = ({DIG}+ "." {DIG}*) | ({DIG}* "." {DIG}+)
DOT_DIG16 = ({DIG16}+ "." {DIG16}*) | ({DIG16}* "." {DIG16}+)
EXP = [eEpP] [+-]? {DIG}+

%state STRING, COMMENT, BLOCKCOMMENT

%%

<YYINITIAL> {
  "auto"                        { String t = yytext(); return new Token(accept(t), t, AUTO, t); }
  "break"                       { String t = yytext(); return new Token(accept(t), t, BREAK, t); }
  "case"                        { String t = yytext(); return new Token(accept(t), t, CASE, t); }
  "char"                        { String t = yytext(); return new Token(accept(t), t, CHAR, t); }
  "const"                       { String t = yytext(); return new Token(accept(t), t, CONST, t); }
  "continue"                    { String t = yytext(); return new Token(accept(t), t, CONTINUE, t); }
  "default"                     { String t = yytext(); return new Token(accept(t), t, DEFAULT, t); }
  "do"                          { String t = yytext(); return new Token(accept(t), t, DO, t); }
  "double"                      { String t = yytext(); return new Token(accept(t), t, DOUBLE, t); }
  "else"                        { String t = yytext(); return new Token(accept(t), t, ELSE, t); }
  "enum"                        { String t = yytext(); return new Token(accept(t), t, ENUM, t); }
  "extern"                      { String t = yytext(); return new Token(accept(t), t, EXTERN, t); }
  "float"                       { String t = yytext(); return new Token(accept(t), t, FLOAT, t); }
  "for"                         { String t = yytext(); return new Token(accept(t), t, FOR, t); }
  "goto"                        { String t = yytext(); return new Token(accept(t), t, GOTO, t); }
  "if"                          { String t = yytext(); return new Token(accept(t), t, IF, t); }
  "int"                         { String t = yytext(); return new Token(accept(t), t, INT, t); }
  "long"                        { String t = yytext(); return new Token(accept(t), t, LONG, t); }
  "register"                    { String t = yytext(); return new Token(accept(t), t, REGISTER, t); }
  "return"                      { String t = yytext(); return new Token(accept(t), t, RETURN, t); }
  "short"                       { String t = yytext(); return new Token(accept(t), t, SHORT, t); }
  "signed"                      { String t = yytext(); return new Token(accept(t), t, SIGNED, t); }
  "sizeof"                      { String t = yytext(); return new Token(accept(t), t, SIZEOF, t); }
  "static"                      { String t = yytext(); return new Token(accept(t), t, STATIC, t); }
  "struct"                      { String t = yytext(); return new Token(accept(t), t, STRUCT, t); }
  "switch"                      { String t = yytext(); return new Token(accept(t), t, SWITCH, t); }
  "typedef"                     { String t = yytext(); return new Token(accept(t), t, TYPEDEF, t); }
  "union"                       { String t = yytext(); return new Token(accept(t), t, UNION, t); }
  "unsigned"                    { String t = yytext(); return new Token(accept(t), t, UNSIGNED, t); }
  "void"                        { String t = yytext(); return new Token(accept(t), t, VOID, t); }
  "volatile"                    { String t = yytext(); return new Token(accept(t), t, VOLATILE, t); }
  "while"                       { String t = yytext(); return new Token(accept(t), t, WHILE, t); }
  
  "..."                         { String t = yytext(); return new Token(accept(t), t, ELLIPSIS, t); }
  "+="                          { String t = yytext(); return new Token(accept(t), t, ADD_EQ, t); }
  "-="                          { String t = yytext(); return new Token(accept(t), t, SUB_EQ, t); }
  "*="                          { String t = yytext(); return new Token(accept(t), t, MUL_EQ, t); }
  "/="                          { String t = yytext(); return new Token(accept(t), t, DIV_EQ, t); }
  "%="                          { String t = yytext(); return new Token(accept(t), t, MOD_EQ, t); }
  "<<="                         { String t = yytext(); return new Token(accept(t), t, LSHIFT_EQ, t); }
  ">>="                         { String t = yytext(); return new Token(accept(t), t, RSHIFT_EQ, t); }
  "&="                          { String t = yytext(); return new Token(accept(t), t, AND_EQ, t); }
  "|="                          { String t = yytext(); return new Token(accept(t), t, OR_EQ, t); }
  "^="                          { String t = yytext(); return new Token(accept(t), t, XOR_EQ, t); }
  "&&"                          { String t = yytext(); return new Token(accept(t), t, ANDAND, t); }
  "||"                          { String t = yytext(); return new Token(accept(t), t, OROR, t); }
  "=="                          { String t = yytext(); return new Token(accept(t), t, EQEQ, t); }
  "!="                          { String t = yytext(); return new Token(accept(t), t, NOTEQ, t); }
  "<="                          { String t = yytext(); return new Token(accept(t), t, LE, t); }
  ">="                          { String t = yytext(); return new Token(accept(t), t, GE, t); }
  "<<"                          { String t = yytext(); return new Token(accept(t), t, LSHIFT, t); }
  ">>"                          { String t = yytext(); return new Token(accept(t), t, RSHIFT, t); }
  "++"                          { String t = yytext(); return new Token(accept(t), t, ADDADD, t); }
  "--"                          { String t = yytext(); return new Token(accept(t), t, SUBSUB, t); }
  "->"                          { String t = yytext(); return new Token(accept(t), t, ARROW, t); }
  
  "__native__"                  { String t = yytext(); return new Token(accept(t), t, NATIVE, t); }
  
  [a-zA-Z_][a-zA-Z0-9_]*        { String t = yytext(); return new Token(accept(t), t, 
                                                              (parser.isTypeName(t) ? TYPENAME : IDENTIFIER), t); }
  (({DOT_DIG} | (({DOT_DIG} | ({DIG}+)) {EXP})) [dDfF]*) |
  ({ZEROX} ({DOT_DIG16} | (({DOT_DIG16} | ({DIG16}+)) {EXP})) [dDfF]*) {
    String t = yytext();
    boolean isFloat = false;
    int len = t.length();
    for (;; len--) {
      char suffix = t.charAt(len - 1);
      if (suffix == 'd' || suffix == 'D' || suffix == 'f' || suffix == 'F') isFloat = true;
      else break;
    }
    Location loc = accept(t);
    try {
      double val = Double.parseDouble(t.substring(0, len));
      return new Token(loc, t, (isFloat ? FLOAT_LITERAL : DOUBLE_LITERAL), val);
    } catch (NumberFormatException e) {
      ExceptionAccumulator.add(loc, "invalidFloatLiteral", t);
      return new Token(loc, t, DOUBLE_LITERAL, 0.0);
    }
  }
  ({DIG}+ [ulUL]*) | ({ZEROX} {DIG16}+ [ulUL]*) {
    String t = yytext();
    boolean isUnsigned = false, isLong = false;
    int len = t.length();
    for (;; len--) {
      char suffix = t.charAt(len - 1);
      if (suffix == 'u' || suffix == 'U') isUnsigned = true;
      else if (suffix == 'l' || suffix == 'L') isLong = true;
      else break;
    }
    Location loc = accept(t);
    try {
      long val = Long.decode(t.substring(0, len));
      if (val < (isUnsigned ? 0 : Integer.MIN_VALUE) || val >= (isUnsigned ? 1L << 32 : 1L << 31))
        throw new NumberFormatException();
      return new Token(loc, t, (isLong ? (isUnsigned ? ULONG_LITERAL : LONG_LITERAL)
                                       : (isUnsigned ? UINT_LITERAL : INT_LITERAL)), val);
    } catch (NumberFormatException e) {
      ExceptionAccumulator.add(loc, "invalidIntegerLiteral", t);
      return new Token(loc, t, INT_LITERAL, 0);
    }
  }
  "'" | "\""                    {
    String t = yytext();
    stringLoc = accept(t); isString = (t.charAt(0) == '"'); string = new StringBuilder(); yybegin(STRING);
  }
  "//"                          { accept(yytext()); yybegin(COMMENT); }
  "/*"                          { accept(yytext()); yybegin(BLOCKCOMMENT); }
  {SP}+                         { accept(yytext()); }
  {EOL}                         { advanceLine(); }
  [\x00-\xFF]                   { String t = yytext(); return new Token(accept(t), t, (int)t.charAt(0)); }
  .                             { String t = yytext(); return new Token(accept(t), t, UNKNOWN); }
  {SP}* "#" {SP}* [0-9]+ {SP}* "\"" [^\"\n\r]* "\"" ({SP} | [0-9])* {EOL} {
    String t = yytext();
    String[] ts = t.substring(0, t.lastIndexOf('"')).trim().split("[ \t]+");
    if (charNumber == 1 && ts.length == 3) {
      try {
        lineNumber = Integer.parseInt(ts[1]) - 1;
        fileName = ts[2].substring(1, ts[2].length());
      } catch (NumberFormatException e) {
      }
    }
    accept(t);
    advanceLine();
  }
}

<STRING> {
  "\\" [0-7]{1,3}               {
    String t = yytext();
    int code = Integer.parseInt(t.substring(1), 8);
    if (code > 255)
      ExceptionAccumulator.add(accept(t), "invalidEscape", t);
    else {
      accept(t);
      string.append((char)code);
    }
  }
  "\\\""                        { accept(yytext()); string.append("\""); }
  "\\\'"                        { accept(yytext()); string.append("'"); }
  "\\n"                         { accept(yytext()); string.append("\n"); }
  "\\r"                         { accept(yytext()); string.append("\r"); }
  "\\t"                         { accept(yytext()); string.append("\t"); }
  "\\\\"                        { accept(yytext()); string.append("\\"); }
  "\\" .                        {
    String t = yytext(); ExceptionAccumulator.add(accept(t), "invalidEscape", t);
  }
  "'" | "\""                    {
    String t = yytext();
    if (isString == (t.charAt(0) == '"')) {
      accept(t); t = string.toString(); string = null;
      yybegin(YYINITIAL); return constructStringToken(stringLoc, t);
    } else {
      accept(t); string.append(t);
    }
  }
  {EOL}                         {
    ExceptionAccumulator.add(accept(yytext()), (isString ? "string" : "char") + "LiteralNotTerminated");
    advanceLine();
    String t = string.toString(); string = null;
    yybegin(YYINITIAL); return constructStringToken(stringLoc, t);
  }
  .                             { String t = yytext(); accept(t); string.append(t); }
}

<COMMENT> {
  {EOL}                         { advanceLine(); yybegin(YYINITIAL); }
  .                             { accept(yytext()); }
}

<BLOCKCOMMENT> {
  {EOL}                         { advanceLine(); }
  "*/"                          { accept(yytext()); yybegin(YYINITIAL); }
  .                             { accept(yytext()); }
}
