package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.comp.ctype.*;
import ctalz.util.*;

public class RelExpr extends Expr
{
	private Op op;
	private Expr lhs, rhs;
	
	public RelExpr(Op o, Expr l, Expr r)
	{
		super(l.loc);
		op = o; lhs = l; rhs = r;
		type = BasicType.INT;
	}
	
	enum Op
	{
		EQ(500, "==", Jump.Op.NE), NE(500, "!=", Jump.Op.EQ), 
		LT(501, "<", Jump.Op.GE), LE(501, "<=", Jump.Op.GT), 
		GT(501, ">", Jump.Op.LE), GE(501, ">=", Jump.Op.LT);
		
		final int priority;
		final String name;
		final Jump.Op jumpOp;
		Op(int p, String n, Jump.Op o) { priority = p; name = n; jumpOp = o; }
	}
	
	int priority()
	{
		return op.priority;
	}
	
	Expr constructEnv(Context ctx, boolean pointerTaken, Env env)
	{
		lhs = lhs.constructEnv(env); rhs = rhs.constructEnv(env);
		CType lt = lhs.type, rt = rhs.type;
		if (lt.isNumber() && rt.isNumber()) {
			Pair<Expr, Expr> p = BinaryExpr.arithmeticConvert(lhs, rhs, op.name.trim());
			lhs = p.fst; rhs = p.snd;
		} else if (CType.areEquivalentPtrs(lt, rt))
			;
		else if ((op == Op.EQ || op == Op.NE) && 
				 ((lt.isPtr() && (rhs instanceof ConstExpr && ((ConstExpr)rhs).isZero())) || 
				  (rt.isPtr() && (lhs instanceof ConstExpr && ((ConstExpr)lhs).isZero())) || 
				  (lt.isPtr() && rt.isPtr() && (lt.isVoidPtr() || rt.isVoidPtr()))))
			;
		else
			ExceptionAccumulator.add(loc, "relIncompatible", lt, rt);
		return this;
	}
	
	public Expr transform(List<VarDef> prepend)
	{
		lhs = lhs.transform(prepend);
		rhs = rhs.transform(prepend);
		return this;
	}
	
	public ConstValue eval()
	{
		reportUnsupportedFeature("constExpr");
		return null;
	}
	
	ActValue compileAsRvalue(Program prog, Function fun)
	{
		return new TernaryExpr(this, new ConstExpr(loc, 1), new ConstExpr(loc, 0))
					   .typeConvert(Context.NORMAL, BasicType.INT).compileAsRvalue(prog, fun);
	}
	void compileAsJump(Program prog, Function fun, ActNode dest)
	{
		ActValue l = lhs.compileAsRvalue(prog, fun), r = rhs.compileAsRvalue(prog, fun);
		fun.add(new Jump(op.jumpOp, lhs.type.isUnsigned(), dest, l.userValue, r.userValue));
	}
	
	public String toString()
	{
		String l = lhs.toString(), r = rhs.toString();
		if (lhs.priority() < priority())
			l = "(" + l + ")";
		if (rhs.priority() <= priority())
			r = "(" + r + ")";
		return l + " " + op.name + " " + r;
	}
}
