package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.comp.ctype.*;
import ctalz.util.*;

public class SwitchStmt extends Stmt
{
	private Expr expr;
	private CompoundStmt body;
	private Label end = new Label("endswitch");
	
	public SwitchStmt(Location l, Expr e, Stmt b)
	{
		super(l);
		expr = e; body = (b instanceof CompoundStmt ? (CompoundStmt)b : 
						  new CompoundStmt(b.loc, Collections.singletonList(b)));
	}
	
	public Stmt constructEnv(Env env)
	{
		env.breakLabels.push(end);
		expr = new CastExpr(expr.loc, BasicType.INT, expr.constructEnv(env));
		body = body.constructEnv(env);
		env.breakLabels.pop();
		return this;
	}
	
	public Stmt transform(List<VarDef> prepend)
	{
		expr = expr.transform(prepend);
		body = body.transform(prepend);
		return this;
	}
	
	public void compile(Program prog, Function init, Function fun)
	{
		ActValue v = expr.compileAsRvalue(prog, fun);
		List<Pair<ConstValue, Label>> labels = new LinkedList<Pair<ConstValue, Label>>();
		labels.add(new Pair<ConstValue, Label>(null, null));
		for (Stmt s : body.stmts)
			collectLabels(prog, fun, s, labels);
		Label def = labels.remove(0).snd;
		if (def == null)
			def = end;
		
		for (Pair<ConstValue, Label> p : labels)
			fun.add(new Jump(Jump.Op.EQ, false, p.snd.node, v.userValue, new Const((int)p.fst.intValue)));
		fun.add(new Jump(def.node));
		
		Iterator<Pair<ConstValue, Label>> it = labels.iterator();
		for (Stmt s : body.stmts)
			compileLabels(prog, init, fun, s, it, def);
		
		fun.add(end.node);
	}
	private void collectLabels(Program prog, Function fun, Stmt s, List<Pair<ConstValue, Label>> labels)
	{
		if (s instanceof LabelStmt) {
			LabelStmt l = (LabelStmt)s;
			if (l.kind == LabelStmt.Kind.CASE)
				labels.add(new Pair<ConstValue, Label>(l.exprVal, new Label("case")));
			else if (l.kind == LabelStmt.Kind.DEFAULT) {
				if (labels.get(0).snd != null)
					ExceptionAccumulator.add(s.loc, "dupDefault");
				else
					labels.get(0).snd = new Label("default");
			}
			collectLabels(prog, fun, l.body, labels);
		} else if (s instanceof CompoundStmt) {
			for (Stmt t : ((CompoundStmt)s).stmts)
				collectLabels(prog, fun, t, labels);
		}
	}
	private void compileLabels(Program prog, Function init, Function fun, 
							   Stmt s, Iterator<Pair<ConstValue, Label>> labels, Label def)
	{
		if (s instanceof LabelStmt) {
			LabelStmt l = (LabelStmt)s;
			if (l.kind == LabelStmt.Kind.CASE)
				fun.add(labels.next().snd.node);
			else if (l.kind == LabelStmt.Kind.DEFAULT)
				fun.add(def.node);
			fun.add(l.label.node);
			compileLabels(prog, init, fun, l.body, labels, def);
		} else if (s instanceof CompoundStmt) {
			for (Stmt t : ((CompoundStmt)s).stmts)
				compileLabels(prog, init, fun, t, labels, def);
		} else
			s.compile(prog, init, fun);
	}
	
	public void writeTo(Writer out, int indent) throws IOException
	{
		out.append(Utilities.tab(indent)).append("switch (").append(expr.toString()).append(')').append(Utilities.EOL);
		body.writeTo(out, indent + 1);
	}
}
