package ctalz.comp.ast;

import java.util.*;
import ctalz.comp.ctype.*;
import ctalz.util.*;

public class TypeSpec extends Spec
{
	final Kind kind;
	private String name;
	private List<VarDef> members;
	private CType ctype;
	
	public enum Kind
	{
		VOID, CHAR, SHORT, INT, LONG, FLOAT, DOUBLE, SIGNED, UNSIGNED, STRUCT, UNION, ENUM, TYPEDEF, CTYPE;
		
		public String toString() { return name().toLowerCase(); }
	}
	
	public TypeSpec(Location l, Kind k)
	{
		super(l);
		kind = k;
	}
	public TypeSpec(Location l, Kind k, String n, List<VarDef> m)
	{
		super(l);
		kind = k; name = n; members = m;
	}
	public TypeSpec(Location l, String n)
	{
		super(l);
		kind = Kind.TYPEDEF; name = n;
	}
	public TypeSpec(Location l, CType t)
	{
		super(l);
		kind = Kind.CTYPE; ctype = t;
	}
	
	public static CType constructCType(Env env, List<TypeSpec> specs)
	{
		CType type = null;
		Spec isShort = null, isLong = null, isSigned = null, isUnsigned = null;
		for (TypeSpec s : specs) {
			switch (s.kind) {
			case VOID: type = checkDup(s.loc, type, BasicType.VOID); break;
			case CHAR: type = checkDup(s.loc, type, BasicType.CHAR); break;
			case SHORT: isShort = checkDup(s.loc, isShort, s); break;
			case INT: type = checkDup(s.loc, type, BasicType.INT); break;
			case LONG: isLong = checkDup(s.loc, isLong, s); break;
			case FLOAT: type = checkDup(s.loc, type, BasicType.FLOAT); break;
			case DOUBLE: type = checkDup(s.loc, type, BasicType.DOUBLE); break;
			case SIGNED: isSigned = checkDup(s.loc, isSigned, s); break;
			case UNSIGNED: isUnsigned = checkDup(s.loc, isUnsigned, s); break;
			case STRUCT: type = checkDup(s.loc, type, s.constructStruct(false, env)); break;
			case UNION: type = checkDup(s.loc, type, s.constructStruct(true, env)); break;
			case ENUM: type = checkDup(s.loc, type, s.constructEnum(env)); break;
			case TYPEDEF: type = s.lookup(env.getEntity(s.loc, s.name)); break;
			case CTYPE: type = s.ctype; break;
			default: assert false;
			}
		}
		if (type == null) type = BasicType.INT;
		if (isShort != null && isLong != null)
			ExceptionAccumulator.add(Utilities.max(isShort.loc, isLong.loc), "specComb");
		else if (isShort != null) {
			assertInt(isShort.loc, type); type = BasicType.SHORT;
		} else if (isLong != null && type != BasicType.DOUBLE) {
			assertInt(isLong.loc, type); type = BasicType.LONG;
		}
		if (isSigned != null && isUnsigned != null)
			ExceptionAccumulator.add(Utilities.max(isSigned.loc, isUnsigned.loc), "specComb");
		else if (isSigned != null)
			assertIntegral(isSigned.loc, type);
		else if (isUnsigned != null)
			type = assertIntegral(isUnsigned.loc, type);
		if (type == BasicType.FLOAT || type == BasicType.DOUBLE) {
			ExceptionAccumulator.addFatal(specs.get(0).loc, 
										  "unsupported", Utilities.localizedString("typeFeature", type));
		}
		return type;
	}
	private static <T> T checkDup(Location loc, T x, T y)
	{
		if (x == null)
			return y;
		else {
			ExceptionAccumulator.add(loc, "dupSpec");
			return x;
		}
	}
	private static void assertInt(Location loc, CType type)
	{
		if (type != BasicType.INT)
			ExceptionAccumulator.add(loc, "specComb");
	}
	private static CType assertIntegral(Location loc, CType type)
	{
		type = (type instanceof BasicType ? ((BasicType)type).unsignedType() : null);
		if (type == null) {
			ExceptionAccumulator.add(loc, "specComb");
			return BasicType.UINT;
		}
		return type;
	}
	private StructType constructStruct(boolean isUnion, Env env)
	{
		if (members == null) {
			StructType result = (isUnion ? env.getUnionOrNull(name) : env.getStructOrNull(name));
			if (result != null)
				return result;
			return StructType.construct(loc, env, isUnion, name, null);
		}
		return StructType.construct(loc, env, isUnion, name, members);
	}
	private CType constructEnum(Env env)
	{
		if (members == null)
			return (CType)lookup(env.getEnum(loc, name));
		return EnumType.construct(loc, env, name, members);
	}
	private CType lookup(Object obj)
	{
		if (obj instanceof CType)
			return (CType)obj;
		else if (obj instanceof Entity && ((Entity)obj).kind == Entity.Kind.TYPE)
			return ((Entity)obj).type;
		ExceptionAccumulator.add(loc, "undefined", "type", name);
		return null;
	}
	
	public String toString()
	{
		if (kind == Kind.STRUCT || kind == Kind.UNION || kind == Kind.ENUM) {
			return (kind + (name != null ? " " + name : "") + 
					(members != null ? " { " + Utilities.enumerate(members, " ") + " }" : ""));
		} else if (kind == Kind.TYPEDEF)
			return name;
		else if (kind == Kind.CTYPE)
			return "<" + ctype + ">";
		else
			return kind.toString();
	}
}
