package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.comp.ctype.*;
import ctalz.util.*;

public class UnaryExpr extends Expr
{
	private Op op;
	private Expr lhs;
	
	public UnaryExpr(Location loc, Op o, Expr l)
	{
		super(loc);
		op = o; lhs = l;
	}
	
	enum Op
	{
		ADDROF("&"), 
		DEREF("*"), 
		PLUS("+"), 
		MINUS("-"), 
		BIT_NOT("~"), 
		NOT("!");
		
		final String name;
		Op(String n) { name = n; }
	}
	
	int priority()
	{
		return 800;
	}
	
	Expr constructEnv(Context ctx, boolean pointerTaken, Env env)
	{
		if (op == Op.NOT)
			return new TernaryExpr(lhs, new ConstExpr(loc, 0), new ConstExpr(loc, 1)).constructEnv(ctx, false, env);
		
		lhs = lhs.constructEnv((op == Op.ADDROF ? Context.KEEP_ARRAY : Context.NORMAL), (op == Op.ADDROF), env);
		if (op == Op.ADDROF)
			return typeConvert(ctx, new PtrType(lhs.type));
		else if (op == Op.DEREF) {
			if (lhs.type.isPtr())
				return typeConvert(ctx, ((PtrType)lhs.type).base);
			else {
				ExceptionAccumulator.add(loc, "nonPointerDeref", lhs.type);
				return typeConvert(ctx, BasicType.INT);
			}
		} else {
			lhs = integralPromote(lhs, op.name.trim());
			if (op == Op.BIT_NOT && !lhs.type.isIntegral())
				ExceptionAccumulator.add(loc, "operandType", op.name.trim(), lhs.type);
			return typeConvert(ctx, lhs.type);
		}
	}
	
	public Expr transform(List<VarDef> prepend)
	{
		if (op == Op.DEREF && type.sizeof() < 0)
			ExceptionAccumulator.add(loc, "sizeNotKnown", type);
		lhs = lhs.transform(prepend);
		return this;
	}
	
	public ConstValue eval()
	{
		if (op == Op.ADDROF || op == Op.DEREF)
			return null;
		ConstValue l = lhs.eval();
		if (l == null)
			return null;
		switch (op) {
		case PLUS: break;
		case MINUS: return l.neg();
		case BIT_NOT: return l.bitNot();
		default: assert false;
		}
		return l;
	}
	
	ActValue compileAsLvalue(Program prog, Function fun)
	{
		if (op != Op.DEREF)
			return super.compileAsLvalue(prog, fun);
		if (hasNoEffects())
			return lhs.compileAsLvalue(prog, fun);
		return lhs.compileAsRvalue(prog, fun);
	}
	ActValue compileAsRvalue(Program prog, Function fun)
	{
		switch (op) {
		case ADDROF:
			if (lhs instanceof UnaryExpr && ((UnaryExpr)lhs).op == Op.DEREF)
				return ((UnaryExpr)lhs).lhs.compileAsRvalue(prog, fun);
			return lhs.compileAsLvalue(prog, fun);
		case DEREF:
			if (hasNoEffects())
				return lhs.compileAsRvalue(prog, fun);
			return lvalueToRvalue(fun, type, compileAsLvalue(prog, fun));
		case PLUS:
			return lhs.compileAsRvalue(prog, fun);
		case MINUS: case BIT_NOT:
			Var u = null;
			ActValue l = lhs.compileAsRvalue(prog, fun);
			fun.add(new Arith((op == Op.MINUS ? Arith.Op.MINUS : Arith.Op.NOT), 
							  lhs.type.isUnsigned(), new VarRef(u = new Var(type)), l.userValue, null));
			return new ActValue(u, (l.baseValue != null ? Const.NULL_BP : null));
		default:
			assert false;
			return null;
		}
	}
	private boolean hasNoEffects()
	{
		CType t = ((PtrType)lhs.type.dequalify()).base.dequalify();
		return (t instanceof ArrayType || t instanceof FunType);
	}
	
	static Expr integralPromote(Expr lhs, String op)
	{
		return BinaryExpr.arithmeticConvert(lhs, lhs, op).fst;
	}
	
	public String toString()
	{
		String l = lhs.toString();
		if (lhs.priority() < priority())
			l = "(" + l + ")";
		return op.name + l;
	}
}
