package ctalz.comp.ast;

import java.util.*;
import ctalz.comp.act.*;
import ctalz.comp.ctype.*;
import ctalz.mac.*;
import ctalz.util.*;

public class VarDef extends ExtDef
{
	public final List<Spec> specs;
	public final List<Decl> decls;
	private List<Entity> entities;
	private boolean isGlobal, isStatic, doTransform;
	
	public VarDef(List<Spec> s, List<Decl> d)
	{
		super(s != null ? s.get(0).loc : d.get(0).loc);
		specs = s; decls = d;
		loc = (specs != null ? specs.get(0).loc : decls.get(0).loc);
	}
	protected VarDef(Location loc, Entity e, Expr i)
	{
		super(loc);
		specs = Collections.singletonList((Spec)new TypeSpec(loc, e.type));
		decls = Collections.singletonList(new Decl(new Decl(loc, e.name), new Init(i)));
		entities = Collections.singletonList(e);
		doTransform = true;
	}
	
	public VarDef constructEnv(Env env)
	{
		CType base = Spec.constructCType(env, specs);
		List<ScSpec> scs = Spec.collectScSpecs(specs);
		boolean typedef = (ScSpec.find(scs, ScSpec.Kind.TYPEDEF) != null);
		entities = new ArrayList<Entity>(decls.size());
		if (ScSpec.find(scs, ScSpec.Kind.STATIC) != null)
			isGlobal = isStatic = true;
		else
			isGlobal = (env.findTopLevel() == null);
		for (Decl d : decls) {
			if (d.kind == Decl.Kind.FUN) {
				if (env.findTopLevel() == null) {
					new FunDef(d.loc, specs, d, null, null, false).constructEnv(env);
					entities.add(null);
				} else
					reportUnsupportedFeature("localFunDecl");
				continue;
			}
			CType t = d.constructCType(env, base);
			String n = d.basemost().name;
			if (isGlobal && !isStatic)
				n = validateName(d.loc, n);
			Entity e = new Entity(typedef ? Entity.Kind.TYPE : Entity.Kind.VAR, n, t);
			env.addEntity(d.loc, n, e);
			entities.add(e);
			if (!typedef && t.dequalify() == BasicType.VOID)
				ExceptionAccumulator.add(d.loc, "voidDecl", n);
			if (isGlobal)
				e.transformDeref = true;
			
			if (d.kind == Decl.Kind.INIT) {
				if (d.init.expr == null)
					reportUnsupportedFeature("aggrInit");
				else {
					d.init.expr = d.init.expr.constructEnv(env);
					if (isGlobal) {
						ConstValue v = d.init.expr.eval();
						if (v == null || !v.type.isIntegral())
							ExceptionAccumulator.add(d.init.expr.loc, "shouldBeIntConst");
					}
				}
			}
		}
		return this;
	}
	static String validateName(Location loc, String name)
	{
		if (name.equals(FunDef.mallocName) || name.equals("GC_malloc") || 
			(Arch.os() == Arch.OSType.GENERAL && name.equals("__main"))) {
			ExceptionAccumulator.add(loc, "reservedName", name);
			return "$error" + Utilities.uniqueID();
		}
		return name;
	}
	
	public VarDef transform(List<VarDef> prepend)
	{
		ListIterator<Decl> it = decls.listIterator();
		for (Entity e : entities) {
			Decl d = it.next();
			if (e != null && e.kind == Entity.Kind.VAR) {
				if (e.type.sizeof() < 0) {
					ExceptionAccumulator.add(d.loc, "sizeNotKnown", e.type);
					continue;
				}
				CType t = transformedType(e);
				if (t != null) {
					String n = e.name;
					VarDef newDef = newMallocDef(d.loc, n, e.type.sizeof(), isGlobal, isStatic, e);
					e.type = t;
					e.transformDeref = true;
					d.basemost().name = (isGlobal ? "(*global " : "(*heapified ") + n + ")";
					if (prepend != null)
						prepend.add(newDef);
				}
				if (d.kind == Decl.Kind.INIT && d.init.expr != null)
					d.init.expr = d.init.expr.transform(prepend);
			}
		}
		return this;
	}
	private static PtrType transformedType(Entity e)
	{
		CType t = e.type.dequalify();
		if (t instanceof ArrayType)
			return new PtrType(((ArrayType)t).base);
		else if (e.isPointerTaken || e.transformDeref || t instanceof StructType)
			return new PtrType(t);
		return null;
	}
	static VarDef newMallocDef(Location loc, String name, int size, boolean isGlobal, boolean isStatic, Entity entity)
	{
		Decl d = new Decl(new Decl(loc, name), 
						  new Init(new CallExpr(loc, FunDef.mallocName, new ConstExpr(loc, size))));
		VarDef v = new VarDef(Collections.singletonList((Spec)new TypeSpec(loc, isGlobal ? "$global" : "$heapified")), 
							  Collections.singletonList(d));
		v.isGlobal = isGlobal;
		v.isStatic = isStatic;
		v.entities = Collections.singletonList(entity);
		return v;
	}
	
	public void compile(Program prog, Function init, Function fun)
	{
		Iterator<Decl> it = decls.iterator();
		for (Entity e : entities) {
			Decl d = it.next();
			if (e != null && e.kind == Entity.Kind.VAR) {
				boolean tr = doTransform;
				if (e.userValue == null) {
					String name = e.name;
					if (isStatic)
						name = ".static" + Utilities.uniqueID() + "_" + name;
					e.userValue = new Var(ActType.fromCType(e.type), name, isGlobal, isStatic, isGlobal);
					if (e.type.hasBaseValue())
						e.baseValue = new Var(ActType.freshPackedBP(), name + ".b", isGlobal, isStatic, isGlobal);
					if (isGlobal) {
						prog.add(new Global(name, false, (isGlobal & !isStatic)));
						if (e.baseValue != null)
							prog.add(new Global(name + ".b", true, (isGlobal & !isStatic)));
					}
				} else
					tr = true;
				if (d.kind == Decl.Kind.INIT) {
					Expr id = new IdentExpr(d.loc, e.name, e, e.type);
					if (tr)
						id = id.transform(null);
					new AssignExpr(AssignExpr.Op.SIMPLE, id, new CastExpr(d.loc, id.type, d.init.expr))
							.compile(prog, (isGlobal ? init : fun));
				}
			}
		}
	}
	
	public String toString()
	{
		return (specs == null ? "" : Utilities.enumerate(specs, " ") + " ") + Utilities.enumerate(decls, ", ") + ";";
	}
}
