package ctalz.tc;

import java.io.*;
import java.util.*;
import ctalz.tc.parser.*;
import ctalz.tc.inst.*;
import ctalz.tc.checker.*;
import ctalz.cnstr.*;
import ctalz.mac.*;
import ctalz.util.*;

public class Main
{
	private static Location fileLoc;
	
	public static void main(String[] args)
	{
		int exitCode = 0;
		long time = System.nanoTime();
		try {
			Utilities.loadResources("ctalz/tc/LocalizedStrings", 
									"ctalz/type/LocalizedStrings", 
									"ctalz/cnstr/LocalizedStrings", 
									"ctalz/util/LocalizedStrings");
			if (args.length < 1 || args.length > 3) {
				System.err.println(Utilities.localizedString("usage"));
				System.exit(1);
			}
			String inputFile = args[0], 
				   outputFile = (args.length >= 2 ? args[1] : Utilities.changeFileSuffix(inputFile, ".cts", ".s")), 
				   encoding = (args.length >= 3 ? args[2] : null);
			
			List<Inst> insts = new Parser(inputFile, encoding).parse();
			if (ExceptionAccumulator.numberOfExceptions() > 0)
				throw new AbortError();
			
			Progress.begin();
			TypeChecker tc = new TypeChecker(insts);
			tc.typecheck();
			if (ExceptionAccumulator.numberOfExceptions() > 0)
				throw new AbortError();
			Progress.end();
			
			try {
				PrintStream out;
				if (outputFile.equals("-"))
					out = System.out;
				else
					out = new PrintStream(outputFile);
				for (Inst i : insts)
					i.writeNative(tc, out);
				emitPostamble(out);
				out.flush();
				if (out.checkError())
					throw new IOException("I/O Error");
				if (!outputFile.equals("-"))
					out.close();
			} catch (IOException e) {
				ExceptionAccumulator.addFatal(Location.NULL, "ioError", e);
			}
		} catch (AbortError e) {
			Progress.end();
			exitCode = failureExitCode();
		}
		
		if (!ExceptionAccumulator.isEmpty())
			ExceptionAccumulator.printlnTo(System.err);
		
		Debug.println(String.format("%s (%.2f sec total; %.2f sec in solver)", 
									(exitCode == 0 ? "Successfully typechecked" : "Typecheck failed"), 
									(System.nanoTime() - time) / 1e9, 
									ConstraintSolver.nanoTimeSpentInSolver() / 1e9));
		System.exit(exitCode);
	}
	
	private static void emitPostamble(PrintStream out) throws IOException
	{
		if (Arch.os() != Arch.OSType.MAC)
			return;
		try {
			for (String line : Utilities.loadResource("/ctalz/tc/postamble"))
				out.println(line);
		} catch (FileNotFoundException e) {
			ExceptionAccumulator.addFatal(Location.NULL, "missingResource", e);
		}
	}
	
	public static Location fileLoc()
	{
		return (fileLoc == null ? Location.NULL : fileLoc);
	}
	public static void setFileLoc(Location loc)
	{
		fileLoc = loc;
	}
	
	public static boolean isReservedLabelName(String name)
	{
		return (name.startsWith(".__ctalz__") || name.equals(exitFunctionName()) || name.equals(mallocFunctionName()));
	}
	public static String uniqueReservedLabelName()
	{
		return ".__ctalz__" + Utilities.uniqueID();
	}
	public static String exitFunctionName()
	{
		return Arch.encodeLabelName("_exit");
	}
	public static String exitFunctionStubName()
	{
		return (Arch.os() == Arch.OSType.MAC ? ".__ctalz____exit" : exitFunctionName());
	}
	public static String mallocFunctionName()
	{
		return Arch.encodeLabelName("GC_malloc");
	}
	public static String mallocFunctionStubName()
	{
		return (Arch.os() == Arch.OSType.MAC ? ".__ctalz___GC_malloc" : mallocFunctionName());
	}
	
	public static int failureExitCode()
	{
		return 1;
	}
}
