package ctalz.tc.checker;

import java.util.*;
import ctalz.type.*;
import ctalz.cnstr.*;
import ctalz.mac.*;
import ctalz.util.*;

public class State
{
	public Type[] reg;
	public List<Pair<Expr, BasicType>> heap;
	public ConstraintSolver solver;
	public static final int tmpRegID = Register.spID + 1;
	
	protected State(ConstraintSolver cs)
	{
		reg = new Type[Register.numberOfGPRs + 2];
		heap = new LinkedList<Pair<Expr, BasicType>>();
		solver = cs;
	}
	protected State copy()
	{
		State result = new State(solver);
		result.reg = reg.clone();
		result.heap = new LinkedList<Pair<Expr, BasicType>>();
		for (Pair<Expr, BasicType> p : heap)
			result.heap.add(p.copy());
		return result;
	}
	
	public Pair<Expr, BasicType> lookupHeap(Location loc, Expr addr) throws TypeCheckException
	{
		Pair<Expr, BasicType> result = lookupHeapOrNull(addr);
		if (result == null)
			throw new TypeCheckException(loc, "lookupHeapFailed", solver.simplify(addr));
		return result;
	}
	public Pair<Expr, BasicType> lookupHeapOrNull(Expr addr)
	{
		for (int i = 0; ; i++) {
			for (Iterator<Pair<Expr, BasicType>> it = heap.iterator(); it.hasNext(); ) {
				Pair<Expr, BasicType> p = it.next();
				if (addr == p.fst || addr.deepEquals(p.fst))
					return p;
			}
			if (i == 1)
				break;
			Expr s = solver.simplify(addr);
			if (addr == s || addr.deepEquals(s))
				break;
			addr = s;
		}
		for (ListIterator<Pair<Expr, BasicType>> it = heap.listIterator(heap.size()); it.hasPrevious(); ) {
			Pair<Expr, BasicType> p = it.previous();
			if (solver.isValid(new RelProp(RelProp.Op.EQ, addr, p.fst)))
				return p;
		}
		return null;
	}
	
	public void checkEntailment(Location loc, RegFileType r, HeapType h, Prop c) throws TypeCheckException
	{
		checkEntailment(loc, r);
		checkEntailment(loc, h);
		checkEntailment(loc, c);
	}
	public void checkEntailment(Location loc, RegFileType r) throws TypeCheckException
	{
		for (int i = 0; i <= Register.numberOfGPRs; i++) {
			if (!reg[i].isEquivalent(r.typeof(i), solver))
				throw new TypeCheckException(loc, "notEntailReg", Register.nameForID(i), reg[i], r.typeof(i));
		}
	}
	public void checkEntailment(Location loc, HeapType h) throws TypeCheckException
	{
		for (UnmodifiablePair<Expr, BasicType> p2 : h.types()) {
			if (p2.snd().isNull(solver))
				continue;
			Pair<Expr, BasicType> p1 = lookupHeapOrNull(p2.fst());
			if (p1 == null)
				throw new TypeCheckException(loc, "notEntailHeapAddr", p2.fst());
			if (!p1.snd.isEquivalent(p2.snd(), solver))
				throw new TypeCheckException(loc, "notEntailHeap", p2.fst(), p1.snd, p2.snd());
		}
	}
	public void checkEntailment(Location loc, Prop c) throws TypeCheckException
	{
		if (!solver.isValid(c))
			throw new TypeCheckException(loc, "notEntailCnstr", c);
	}
	
	public static String registerNameForID(int index)
	{
		return registerNameForID(index, Arch.wordSize);
	}
	public static String registerNameForID(int index, int size)
	{
		if (index == State.tmpRegID)
			return Utilities.localizedString("tmpRegName");
		else
			return Register.nameForID(index, size);
	}
	
	public String toString()
	{
		StringBuilder result = new StringBuilder();
		result.append("Registers:");
		for (int i = 0; i < reg.length; i++)
			result.append(Utilities.EOL).append("  ").append(registerNameForID(i)).append(" = ").append(reg[i]);
		result.append(Utilities.EOL).append("Heap:");
		for (Pair<Expr, BasicType> p : heap)
			result.append(Utilities.EOL).append("  ").append(p.fst).append(" = ").append(p.snd);
		return result.toString();
	}
}
