package ctalz.tc.checker;

import java.util.*;
import ctalz.tc.*;
import ctalz.tc.inst.*;
import ctalz.type.*;
import ctalz.cnstr.*;
import ctalz.mac.*;
import ctalz.util.*;

public class TypeChecker
{
	private List<Inst> insts;
	private Map<String, BasicType> labelAddrs = new HashMap<String, BasicType>();
	private Map<String, BasicType> labelTypes = new HashMap<String, BasicType>();
	private List<Pair<Expr, BasicType>> initHeap;
	public State state;
	public boolean suppressUnreachableWarnings;
	public static boolean warnUnreachable = (Config.asInt("WarnUnreachableInstructions", 0) != 0);
	public static boolean warnUnsafe = (Config.asInt("WarnUnsafeInstructions", 1) != 0);
	
	public TypeChecker(List<Inst> i)
	{
		insts = i;
	}
	
	public void typecheck()
	{
		List<Inst> entries = new LinkedList<Inst>();
		for (Iterator<Inst> it = insts.iterator(); it.hasNext(); ) {
			Inst i = it.next();
			if (i instanceof EntryInst) {
				entries.add(i);
				it.remove();
			} else if (i instanceof CommentInst)
				it.remove();
		}
		if (entries.isEmpty())
			ExceptionAccumulator.addFatal(Main.fileLoc(), "noEntry");
		
		collectLabels();
		
		boolean skip = false;
		state = new State(ConstraintSolver.newInstance());
		initState();
		insts.addAll(0, entries);
		try {
			Progress.setMaximum(insts.size());
			for (ListIterator<Inst> it = insts.listIterator(); it.hasNext(); ) {
				Inst i = it.next();
				try {
					if (i instanceof LabelInst)
						skip = false;
					else if (skip) {
						Progress.advance();
						continue;
					}
					
					Debug.fillLine('=');
					Debug.println("Instruction at " + i.loc() + ":");
					Debug.println("  " + i);
					Debug.println(state);
					Debug.fillLine('-');
					
					i.typecheck(this, it);
					
					Debug.println(state);
				} catch (TypeCheckException e) {
					ExceptionAccumulator.add(e.loc(), e.message());
					state.solver.assume(ConstProp.FALSE);
					skip = true;
				}
				Progress.advance();
			}
			Progress.set(1, 1, "");
		} finally {
			state.solver.terminate();
		}
	}
	
	private void collectLabels()
	{
		initHeap = new LinkedList<Pair<Expr, BasicType>>();
		for (Inst i : insts) {
			if (!(i instanceof LabelInst))
				continue;
			LabelInst label = (LabelInst)i;
			String name = label.name();
			if (labelAddrs.containsKey(name))
				ExceptionAccumulator.add(label.loc(), "duplicateLabel", name);
			else {
				labelAddrs.put(name, new IntType(label.addr()));
				labelTypes.put(name, label.type());
				initHeap.add(new Pair<Expr, BasicType>(label.addr(), label.type()));
			}
		}
	}
	public BasicType typeOfLabel(Location loc, String name) throws TypeCheckException
	{
		BasicType t = labelAddrs.get(name);
		if (t == null)
			throw new TypeCheckException(loc, "undefinedLabel", name);
		return t;
	}
	public BasicType typeOfLabeledMemory(Location loc, String name) throws TypeCheckException
	{
		BasicType t = labelTypes.get(name);
		if (t == null)
			throw new TypeCheckException(loc, "undefinedLabel", name);
		return t;
	}
	
	public static VarExpr freshIndexVar()
	{
		return VarExpr.freshInstance();
	}
	
	public void initState()
	{
		initRegs();
		initHeap();
		state.solver.reset();
	}
	private void initRegs()
	{
		for (int i = 0; i < Register.numberOfGPRs; i++)
			state.reg[i] = new IntType(TypeChecker.freshIndexVar());
		state.reg[Register.spID] = new StackType(Location.NULL, Collections.singletonList(new TypeVar("<?>")));
		state.reg[state.tmpRegID] = new IntType(TypeChecker.freshIndexVar());
	}
	private void initHeap()
	{
		state.heap.clear();
		for (Pair<Expr, BasicType> p : initHeap)
			state.heap.add(p.copy());
	}
	public void initStateTo(CodeType init)
	{
		Type[] initRegs = init.reg().types();
		System.arraycopy(initRegs, 0, state.reg, 0, initRegs.length);
		state.reg[state.tmpRegID] = new IntType(TypeChecker.freshIndexVar());
		
		initHeap();
		for (UnmodifiablePair<Expr, BasicType> p : init.heap().types())
			state.heap.add(new Pair<Expr, BasicType>(p.fst(), p.snd()));
		
		state.solver.reset();
		state.solver.assume(init.cnstr());
	}
	
	public boolean isReachable()
	{
		return !state.solver.isValid(ConstProp.FALSE);
	}
	public void checkReachability(Location loc)
	{
		if (warnUnreachable && !suppressUnreachableWarnings && !isReachable()) {
			ExceptionAccumulator.addWarning(loc, "unreachable");
			suppressUnreachableWarnings = true;
		}
	}
	
	public void checkTypeProof(Location loc, 
							   Type init, TupleType type, List<List<Inst>> proof) throws TypeCheckException
	{
		int count = proof.size();
		if (count != type.length())
			throw new TypeCheckException(loc, "invalidProofSize", proof.size(), type.length(), type);
		for (int i = 0; i < count; i++) {
			Type t = type.element(i);
			if (t instanceof ArrayType)
				t = ((ArrayType)t).elementType();
			checkTypeProof((count == 1 ? -1 : i), loc, init, (t != null ? t : init), proof.get(i));
		}
	}
	private void checkTypeProof(int index, Location loc, 
								Type init, Type type, List<Inst> proof) throws TypeCheckException
	{
		State savedState = state;
		state = state.copy();
		state.solver.saveEnvironment();
		initRegs();
		state.reg[0] = init;
		for (Inst i : proof) {
			if (!(i instanceof CommentInst) && 
				!(i instanceof PackInst) && !(i instanceof GuardInst) && !(i instanceof RollInst) && 
				!(i instanceof ApplyInst))
				throw new TypeCheckException(i.loc(), "invalidProofInst");
			i.typecheck(this, null);
		}
		if (!state.reg[0].isEquivalent(type, state.solver)) {
			if (index >= 0)
				throw new TypeCheckException(loc, "invalidProofAt", index, state.reg[0], type);
			else
				throw new TypeCheckException(loc, "invalidProof", state.reg[0], type);
		}
		state.solver.restoreEnvironment();
		state = savedState;
	}
}
