package ctalz.tc.inst;

import java.io.*;
import java.util.*;
import ctalz.tc.checker.*;
import ctalz.type.*;
import ctalz.cnstr.*;
import ctalz.mac.*;
import ctalz.util.*;

public class ArithInst extends Inst
{
	private Op op;
	private Operand left, right;
	
	public static enum Op
	{
		ADD(OpExpr.Op.ADD, 2, "addl"), SUB(OpExpr.Op.SUB, 2, "subl"), 
		MUL(OpExpr.Op.MUL, 2, "imull"), DIV(OpExpr.Op.DIV, 2, "idivl"), MOD(null, 2, "idivl"), 
		UMUL(OpExpr.Op.MUL, 2, "mull"), UDIV(OpExpr.Op.DIV, 2, "divl"), UMOD(null, 2, "divl"), 
		LSL(null, 2, "shll"), LSR(null, 2, "shrl"), ASR(null, 2, "sarl"), 
		SE1(null, 2, "movsbl", Arch.wordSize, 1), SE2(null, 2, "movswl", Arch.wordSize, 2), 
		ZE1(null, 2, "movzbl", Arch.wordSize, 1), ZE2(null, 2, "movzwl", Arch.wordSize, 2), 
		AND(null, 2, "andl"), OR(null, 2, "orl"), XOR(null, 2, "xorl"), NOT(null, 1, "notl"), 
		NEG(OpExpr.Op.NEG, 1, "negl"), 
		MOV(null, 2, "movl");
		
		private int arity;
		private OpExpr.Op indexOp;
		private String mnemonic;
		private int leftSize, rightSize;
		Op(OpExpr.Op o, int a, String m) {
			indexOp = o; arity = a; mnemonic = m; leftSize = rightSize = Arch.wordSize;
		}
		Op(OpExpr.Op o, int a, String m, int l, int r) {
			indexOp = o; arity = a; mnemonic = m; leftSize = l; rightSize = r;
		}
	}
	
	public ArithInst(Location loc, Op o, Operand l)
	{
		super(loc);
		assert o.arity == 1;
		op = o; left = l;
	}
	public ArithInst(Location loc, Op o, Operand l, Operand r)
	{
		super(loc);
		assert o.arity == 2;
		op = o; left = l; right = r;
	}
	
	public void typecheck(TypeChecker tc, ListIterator<Inst> insts) throws TypeCheckException
	{
		tc.checkReachability(loc);
		
		assertOperandKind(left, Operand.Kind.REG);
		assertNotSp(left);
		if (right != null)
			assertOperandKind(right, Operand.Kind.INT, Operand.Kind.LABEL, Operand.Kind.REG);
		
		if (op == Op.MOV) {
			Type rt = right.type(tc);
			assertTypeNotLinear(right.loc, tc, rt);
			left.setType(tc, rt);
		} else {
			IntType lt = assertType(tc, left, IntType.class);
			if (right != null) {
				IntType rt = assertType(tc, right, IntType.class);
				if (op.indexOp == null)
					left.setType(tc, new IntType(TypeChecker.freshIndexVar()));
				else
					left.setType(tc, new IntType(OpExpr.calculate(op.indexOp, lt.value(), rt.value())));
			} else {
				if (op.indexOp == null)
					left.setType(tc, new IntType(TypeChecker.freshIndexVar()));
				else
					left.setType(tc, new IntType(OpExpr.calculate(op.indexOp, lt.value())));
			}
		}
	}
	
	public void writeNative(TypeChecker tc, PrintStream out)
	{
		String ls = left.toString(op.leftSize), rs = (right != null ? right.toString(op.rightSize) : ""), lss = ls;
		boolean destIsEax = ls.equals("%eax"), destIsEcx = ls.equals("%ecx"), destIsEdx = ls.equals("%edx"), 
				rightIsEax = rs.equals("%eax"), rightIsEcx = rs.equals("%ecx");
		switch (op) {
		case DIV:
		case UMUL: case UDIV:
		case MOD: case UMOD:
			if (!destIsEcx) out.println("\tpushl %ecx");
			if (!destIsEax) out.println("\tpushl %eax");
			if (!destIsEdx) out.println("\tpushl %edx");
			if (!rightIsEax) {
				if (!destIsEax)  out.println("\tmovl " + ls + ", %eax");
				if (!rightIsEcx) out.println("\tmovl " + rs + ", %ecx");
			} else if (!destIsEcx) {
				if (!rightIsEcx) out.println("\tmovl " + rs + ", %ecx");
				if (!destIsEax)  out.println("\tmovl " + ls + ", %eax");
			} else {
				out.println("\tmovl %eax, %edx");
				out.println("\tmovl %ecx, %eax");
				out.println("\tmovl %edx, %ecx");
			}
			out.println("\txorl %edx, %edx");
			out.println("\t" + op.mnemonic + " %ecx");
			if (!destIsEdx) {
				if (op == Op.MOD || op == Op.UMOD)
					out.println("\tmovl " + "%edx, " + ls);
				out.println("\tpopl %edx");
			}
			if (!destIsEax) {
				if (op == Op.DIV || op == Op.UMUL || op == Op.UDIV)
					out.println("\tmovl " + "%eax, " + ls);
				out.println("\tpopl %eax");
			}
			if (!destIsEcx)
				out.println("\tpopl %ecx");
			return;
		case LSL: case LSR: case ASR:
			if (destIsEcx) {
				out.println("\tpushl %eax");
				out.println("\tmovl " + ls + ", %eax");
				lss = "%eax";
			}
			if (!rightIsEcx) {
				if (destIsEcx)
					out.println("\tmovl " + (rightIsEax ? "0(%esp)" : rs) + ", %ecx");
				else {
					out.println("\tpushl %ecx");
					out.println("\tmovl " + rs + ", %ecx");
				}
			}
			out.println("\t" + op.mnemonic + " %cl, " + lss);
			if (!rightIsEcx && !destIsEcx)
				out.println("\tpopl %ecx");
			if (destIsEcx) {
				out.println("\tmovl " + lss + ", %ecx");
				out.println("\tpopl " + lss);
			}
			return;
		}
		if (right == null)
			out.println("\t" + op.mnemonic + " " + ls);
		else
			out.println("\t" + op.mnemonic + " " + rs + ", " + ls);
	}
	
	public String toString()
	{
		if (right == null)
			return op.name().toLowerCase() + " " + left;
		else
			return op.name().toLowerCase() + " " + left + ", " + right;
	}
}
