package ctalz.tc.inst;

import java.io.*;
import java.util.*;
import ctalz.tc.checker.*;
import ctalz.type.*;
import ctalz.cnstr.*;
import ctalz.mac.*;
import ctalz.util.*;

public class JmpInst extends Inst
{
	private Cond cond;
	private Operand left, right, label;
	private int divisor;
	
	public static enum Cond
	{
		EQ(RelProp.Op.EQ, 2, "e"), NE(RelProp.Op.NE, 2, "ne"), 
		LT(RelProp.Op.LT, 2, "g"), LE(RelProp.Op.LE, 2, "ge"), 
		GT(RelProp.Op.GT, 2, "l"), GE(RelProp.Op.GE, 2, "le"), 
		LTU(null, 2, "a"), LEU(null, 2, "ae"), 
		GTU(null, 2, "b"), GEU(null, 2, "be"), 
		DV(RelProp.Op.DV, 2, null), ND(RelProp.Op.ND, 2, null), 
		ALWAYS(null, 0, null);
		
		private int arity;
		private RelProp.Op propOp;
		private String nativeName;
		Cond(RelProp.Op o, int a, String n) { propOp = o; arity = a; nativeName = n; }
	}
	
	public JmpInst(Location loc, Cond c, Operand j)
	{
		super(loc);
		assert c.arity == 0;
		cond = c; label = j;
	}
	public JmpInst(Location loc, Cond c, Operand l, Operand r, Operand j)
	{
		super(loc);
		assert c.arity == 2;
		cond = c; left = l; right = r; label = j;
	}
	
	public void typecheck(TypeChecker tc, ListIterator<Inst> insts) throws TypeCheckException
	{
		tc.checkReachability(loc);
		
		if (left != null) {
			assertOperandKind(left, Operand.Kind.INT, Operand.Kind.REG);
			assertOperandKind(right, Operand.Kind.INT, Operand.Kind.REG);
		}
		assertOperandKind(label, Operand.Kind.REG, Operand.Kind.LABEL);
		
		IntType lt = null, rt = null, ct;
		Pair<Expr, BasicType> cht;
		if (left != null) {
			lt = assertType(tc, left, IntType.class);
			rt = assertType(tc, right, IntType.class);
		}
		ct = assertType(tc, label, IntType.class);
		cht = tc.state.lookupHeap(label.loc, ct.value());
		
		CodeType code = assertType(label.loc, tc, cht.snd, CodeType.class);
		if (!code.types().isEmpty() || !code.types().isEmpty())
			throw new TypeCheckException(label.loc, "genericJump", simplify(tc, code));
		if (left != null) {
			// Conditional jump
			Expr lv = lt.value(), rv = rt.value();
			RelProp prop = null;
			if (cond.propOp != null) {
				prop = normalizeProp(RelProp.newOrNull(cond.propOp, lv, rv));
				if (prop == null) {
					lv = tc.state.solver.simplify(lv);
					rv = tc.state.solver.simplify(rv);
					prop = normalizeProp(RelProp.newOrNull(cond.propOp, lv, rv));
				}
			}
			tc.state.solver.saveEnvironment();
			try {
				if (prop != null) {
					if (tc.warnUnreachable) {
						if (tc.state.solver.isValid(prop))
							ExceptionAccumulator.addWarning(loc, "branchTaken");
						else if (tc.state.solver.isUnsatisfiable(prop))
							ExceptionAccumulator.addWarning(loc, "branchNotTaken");
					}
					tc.state.solver.assume(prop);
				} else if (cond.propOp != null)
					ExceptionAccumulator.addWarning(loc, "cannotHandleProp", lv + " " + cond.propOp + " " + rv);
				tc.state.checkEntailment(loc, code.reg(), code.heap(), code.cnstr());
			} finally {
				tc.state.solver.restoreEnvironment();
			}
			if (prop != null)
				tc.state.solver.assume(new CombProp(CombProp.Op.NOT, prop));
		} else {
			// Unconditional jump
			tc.state.checkEntailment(loc, code.reg(), code.heap(), code.cnstr());
			tc.state.solver.assume(ConstProp.FALSE);
		}
	}
	private RelProp normalizeProp(RelProp prop)
	{
		if (prop == null)
			return null;
		if (prop.op() == RelProp.Op.DV || prop.op() == RelProp.Op.ND) {
			if (!(prop.left() instanceof ConstExpr))
				return null;
			divisor = ((ConstExpr)prop.left()).value();
			if (divisor <= 0 || Integer.bitCount(divisor) != 1)
				return null;
		}
		return prop;
	}
	
	public void writeNative(TypeChecker tc, PrintStream out)
	{
		String labelStr = (label.kind() == Operand.Kind.REG ? "*" + label : label.toString(true));
		if (cond == Cond.ALWAYS)
			out.println("\tjmp " + labelStr);
		else if (cond == Cond.DV || cond == Cond.ND) {
			String tmp = (right.kind() == Operand.Kind.REG ? right.toString() : "%eax");
			out.println("\tpushl " + tmp);
			if (right.kind() != Operand.Kind.REG)
				out.println("\tmovl " + right.toString() + ", " + tmp);
			out.println("\tandl $" + (divisor - 1) + ", " + tmp);
			out.println("\tpopl " + tmp);
			if (cond == Cond.DV)
				out.println("\tjeq " + labelStr);
			else
				out.println("\tjne " + labelStr);
		} else {
			out.println("\tcmpl " + left.toString() + ", " + right.toString());
			out.println("\tj" + cond.nativeName.toLowerCase() + " " + labelStr);
		}
	}
	
	public String toString()
	{
		String result;
		if (cond == Cond.ALWAYS)
			return "jmp " + label;
		else
			return "b" + cond.name().toLowerCase() + " " + left + ", " + right + ", " + label;
	}
}
