package ctalz.tc.inst;

import java.io.*;
import java.util.*;
import ctalz.tc.checker.*;
import ctalz.type.*;
import ctalz.cnstr.*;
import ctalz.mac.*;
import ctalz.util.*;

public class MemInst extends Inst
{
	private Op op;
	private Operand reg, addr;
	private int size, elem;
	
	public static enum Op
	{
		LD, ST, LST;
	}
	
	public MemInst(Location loc, Op o, int s, int e, Operand r, Operand a)
	{
		super(loc);
		op = o; size = s; elem = e; reg = r; addr = a;
	}
	
	public void typecheck(TypeChecker tc, ListIterator<Inst> insts) throws TypeCheckException
	{
		tc.checkReachability(loc);
		
		if (!Arch.isAccessibleMemorySize(size))
			throw new TypeCheckException(loc, "memSize", size);
		if (op == Op.LD) {
			assertOperandKind(reg, Operand.Kind.REG);
			assertNotSp(reg);
		} else
			assertOperandKind(reg, Operand.Kind.REG, Operand.Kind.INT, Operand.Kind.LABEL);
		assertOperandKind(addr, Operand.Kind.LABEL, Operand.Kind.DISP_INT, Operand.Kind.DISP_REG);
		
		if (op != Op.LST) {
			Pair<Expr, BasicType> ht = tc.state.lookupHeap(addr.loc, assertType(tc, addr, IntType.class).value());
			Expr ofs = assertType(addr.loc, tc, addr.offsetType(tc), IntType.class).value();
			BasicType t = process(tc, assertType(addr.loc, tc, ht.snd, TupleType.class), ofs);
			if (op == Op.LD)
				reg.setType(tc, t);
		} else {
			LinearType lt = assertType(tc, addr, LinearType.class);
			Expr ofs = assertType(addr.loc, tc, addr.offsetType(tc), IntType.class).value();
			BasicType t = process(tc, lt.body(), ofs);
			addr.setType(tc, new LinearType((TupleType)t));
		}
	}
	private BasicType process(TypeChecker tc, TupleType tuple, Expr offset) throws TypeCheckException
	{
		if (elem < 0 || elem >= tuple.length())
			throw new TypeCheckException(loc, "memElem", elem, 0, tuple.length() - 1, tuple);
		
		ArrayType array = tuple.element(elem);
		Pair<Expr, Expr> arrayOffsetAndSize = checkAccess(tc, tuple, array, offset);
		Expr eltSize = (array.elementType() != null ? 
						array.elementType().sizeof(tc.state.solver) : ConstExpr.valueOf(1));
		checkAlign(tc, array, OpExpr.calculate(OpExpr.Op.SUB, offset, arrayOffsetAndSize.fst), eltSize);
		switch (op) {
		case LD:
			if (array.elementType() != null) {
				checkSize(tc, array, eltSize);
				return array.elementType();
			} else
				return new IntType(TypeChecker.freshIndexVar());
		case ST:
			if (array.elementType() != null) {
				checkSize(tc, array, eltSize);
				return null;
			} else
				return null;
		case LST:
			checkWholeSize(tc, array, arrayOffsetAndSize.snd);
			assertTypeReg(reg.loc, tc, reg.type(tc));
			return tuple.updateElement(elem, 
									  new ArrayType(loc, (BasicType)reg.type(tc).alphaCopy(), ConstExpr.valueOf(size)));
		default:
			assert false;
			return null;
		}
	}
	private Pair<Expr, Expr> checkAccess(TypeChecker tc, 
										 TupleType tuple, ArrayType array, Expr offset) throws TypeCheckException
	{
		Expr arrayOffset = tuple.offsetOf(tc.state.solver, elem), arraySize = array.sizeof(tc.state.solver);
		if (arrayOffset == null || arraySize == null)
			throw new TypeCheckException(loc, "memIndex", offset, simplify(tc, arrayOffset), simplify(tc, (Expr)null));
		
		Expr begin = arrayOffset;
		Expr end = OpExpr.calculate(OpExpr.Op.SUB, 
									OpExpr.calculate(OpExpr.Op.ADD, arrayOffset, arraySize), ConstExpr.valueOf(size));
		if (!tc.state.solver.isValid(new CombProp(CombProp.Op.AND, 
												  new RelProp(RelProp.Op.LE, begin, offset), 
												  new RelProp(RelProp.Op.LE, offset, end)))) {
			throw new TypeCheckException(addr.loc, "memIndex", simplify(tc, offset), 
										 simplify(tc, begin), simplify(tc, end));
		}
		return new Pair<Expr, Expr>(arrayOffset, arraySize);
	}
	private void checkAlign(TypeChecker tc, ArrayType array, Expr offset, Expr align) throws TypeCheckException
	{
		if (align == null) {
			throw new TypeCheckException(addr.loc, "memMisalign", 
										 simplify(tc, offset), simplify(tc, align), simplify(tc, array));
		}
		if (align == ConstExpr.valueOf(1))
			return;
		align = tc.state.solver.simplify(align);
		RelProp prop = RelProp.newOrNull(RelProp.Op.DV, align, offset);
		if (prop == null) {
			offset = tc.state.solver.simplify(offset);
			prop = RelProp.newOrNull(RelProp.Op.DV, align, offset);
		}
		if (prop == null || !tc.state.solver.isValid(prop)) {
			throw new TypeCheckException(addr.loc, "memMisalign", 
										 simplify(tc, offset), simplify(tc, align), simplify(tc, array));
		}
	}
	private void checkSize(TypeChecker tc, ArrayType array, Expr eltSize) throws TypeCheckException
	{
		if (eltSize == null || !tc.state.solver.isEquivalent(eltSize, ConstExpr.valueOf(size)))
			throw new TypeCheckException(loc, "memElemSize", size, simplify(tc, eltSize), array);
	}
	private void checkWholeSize(TypeChecker tc, ArrayType array, Expr arraySize) throws TypeCheckException
	{
		if (!tc.state.solver.isEquivalent(arraySize, ConstExpr.valueOf(size)))
			throw new TypeCheckException(loc, "memElemSize", size, simplify(tc, arraySize), array);
	}
	
	public void writeNative(TypeChecker tc, PrintStream out)
	{
		String suffix = Arch.instructionSuffixForSize(size), a;
		if (addr.kind() == Operand.Kind.LABEL)
			a = "(" + addr.toString(true) + ")";
		else
			a = addr.toString(size);
		if (op == Op.LD)
			out.println("\tmov" + suffix + " " + a + ", " + reg.toString(size));
		else
			out.println("\tmov" + suffix + " " + reg.toString(size) + ", " + a);
	}
	
	public String toString()
	{
		return op.name().toLowerCase() + "<" + size + "," + elem + "> " + reg + ", " + addr;
	}
}
