package ctalz.tc.inst;

import java.io.*;
import java.util.*;
import ctalz.tc.checker.*;
import ctalz.type.*;
import ctalz.cnstr.*;
import ctalz.mac.*;
import ctalz.util.*;

public class StackInst extends Inst
{
	private Op op;
	private Operand allocSize;
	private List<Integer> allocSizes;
	private List<VarExpr> allocAs;
	private int accessSize, elem;
	private Operand reg, addr;
	
	public static enum Op
	{
		SALLOC, SFREE, SLD, SST;
	}
	
	public StackInst(Location loc, List<Integer> ss, Operand s, List<VarExpr> a)
	{
		super(loc);
		op = Op.SALLOC; allocSizes = ss; allocSize = s; allocAs = a;
	}
	public StackInst(Location loc, int e, Operand s)
	{
		super(loc);
		op = Op.SFREE; elem = e; allocSize = s;
	}
	public StackInst(Location loc, Op o, int s, int e, Operand r, Operand a)
	{
		super(loc);
		op = o; accessSize = s; elem = e; reg = r; addr = a;
	}
	
	public void typecheck(TypeChecker tc, ListIterator<Inst> insts) throws TypeCheckException
	{
		tc.checkReachability(loc);
		
		if (allocSize != null)
			assertOperandKind(allocSize, Operand.Kind.REG, Operand.Kind.INT);
		else {
			if (!Arch.isAccessibleMemorySize(accessSize))
				throw new TypeCheckException(loc, "memSize", accessSize);
			if (op == Op.SLD) {
				assertOperandKind(reg, Operand.Kind.REG);
				assertNotSp(reg);
			} else
				assertOperandKind(reg, Operand.Kind.REG, Operand.Kind.INT, Operand.Kind.LABEL);
			assertOperandKind(addr, Operand.Kind.REG, Operand.Kind.INT);
		}
		
		StackType st = (StackType)tc.state.reg[Register.spID];
		if (op == Op.SALLOC) {
			if (allocSizes.size() != allocAs.size())
				throw new TypeCheckException(loc, "sallocElem", allocAs.size(), allocSizes.size());
			List<BasicType> ts = new ArrayList<BasicType>(allocSizes.size());
			Iterator<VarExpr> as = allocAs.iterator();
			int totalSize = 0;
			for (int i : allocSizes) {
				if (i != Arch.wordSize)
					throw new TypeCheckException(loc, "sallocSize", i);
				ts.add(new IntType(as.next()));
				totalSize += i;
			}
			Expr s = assertType(tc, allocSize, IntType.class).value();
			if (!tc.state.solver.isEquivalent(s, ConstExpr.valueOf(totalSize)))
				throw new TypeCheckException(allocSize.loc, "mallocSize", simplify(tc, s), totalSize);
			tc.state.reg[Register.spID] = ((StackType)tc.state.reg[Register.spID]).append(ts);
		} else if (op == Op.SFREE) {
			if (elem <= 0 || elem > st.length())
				throw new TypeCheckException(loc, "sfreeElem", 1, st.length(), simplify(tc, st));
			Expr s = assertType(tc, allocSize, IntType.class).value();
			Expr s2 = st.offsetOf(tc.state.solver, elem);
			if (s2 == null || !tc.state.solver.isEquivalent(s, s2)) {
				throw new TypeCheckException(allocSize.loc, "sfreeSize", 
											 simplify(tc, s), simplify(tc, s2), simplify(tc, st));
			}
			tc.state.reg[Register.spID] = st.remove(elem);
		} else {
			if (elem < 0 || elem >= st.length())
				throw new TypeCheckException(loc, "stackElem", elem, 0, st.length() - 1, simplify(tc, st));
			Expr offset = assertType(tc, addr, IntType.class).value();
			Expr expectedOffset = st.offsetOf(tc.state.solver, elem);
			if (expectedOffset == null || !tc.state.solver.isEquivalent(offset, expectedOffset)) {
				throw new TypeCheckException(addr.loc, "stackOffset", 
											 simplify(tc, offset), simplify(tc, expectedOffset), simplify(tc, st));
			}
			BasicType target = st.element(elem);
			Expr targetSize = target.sizeof(tc.state.solver);
			if (targetSize == null || !tc.state.solver.isEquivalent(ConstExpr.valueOf(accessSize), targetSize)) {
				throw new TypeCheckException(loc, "stackSize", 
											 accessSize, simplify(tc, targetSize), simplify(tc, target));
			}
			if (op == Op.SLD)
				reg.setType(tc, target.alphaCopy());
			else {
				assertTypeNotLinear(reg.loc, tc, reg.type(tc));
				tc.state.reg[Register.spID] = st.updateElement(elem, (BasicType)reg.type(tc).alphaCopy());
			}
		}
	}
	
	public void writeNative(TypeChecker tc, PrintStream out)
	{
		String suffix = Arch.instructionSuffixForSize(accessSize);
		String addrStr = (addr != null ? addr.toString().replace("$", "") + "(%esp)" : null);
		switch (op) {
		case SALLOC:
			out.println("\tsubl " + allocSize.toString() + ", %esp");
			break;
		case SFREE:
			out.println("\taddl " + allocSize.toString() + ", %esp");
			break;
		case SLD:
			out.println("\tmov" + suffix + " " + addrStr + ", " + reg.toString(accessSize));
			break;
		case SST:
			out.println("\tmov" + suffix + " " + reg.toString(accessSize) + ", " + addrStr);
			break;
		default:
			assert false;
		}
	}
	
	public String toString()
	{
		if (op == Op.SALLOC) {
			return ("salloc<" + Utilities.enumerate(allocSizes, ",") + "> " + allocSize + 
					", [" + Utilities.enumerate(allocAs, ",") + "]");
		} else if (op == Op.SFREE)
			return op.name().toLowerCase() + " " + allocSize;
		else
			return op.name().toLowerCase() + "<" + accessSize + "," + elem + "> " + reg + ", " + addr;
	}
}
