package ctalz.util;

import java.io.*;
import java.math.*;
import java.security.*;
import java.util.*;

public class Utilities
{
	public static final String EOL = System.getProperty("line.separator");
	
	private static ResourceBundle bundles[] = new ResourceBundle[0];
	private static String tabString;
	private static int nextID;
	
	private Utilities()
	{
	}
	
	public static void loadResources(String... files)
	{
		bundles = new ResourceBundle[files.length];
		for (int i = 0; i < bundles.length; i++) {
			try {
				bundles[i] = ResourceBundle.getBundle(files[i], Locale.getDefault());
			} catch (MissingResourceException e) {
				try {
					bundles[i] = ResourceBundle.getBundle(files[i], new Locale("en"));
				} catch (MissingResourceException e2) {
					bundles = new ResourceBundle[0];
					ExceptionAccumulator.addFatal(null, "Resource file not found: " + files[i]);
				}
			}
		}
		tabString = localizedStringOrNull("tab");
		if (tabString == null)
			tabString = "  ";
	}
	public static String[] loadResource(String file) throws IOException
	{
		InputStream in = new Object().getClass().getResourceAsStream(file);
		if (in == null)
			throw new IOException("File not found: " + file);
		BufferedReader br = new BufferedReader(new InputStreamReader(in));
		List<String> lines = new ArrayList<String>();
		for (;;) {
			String line = br.readLine();
			if (line == null)
				break;
			lines.add(line);
		}
		return lines.toArray(new String[0]);
	}
	
	public static String callerClass()
	{
		try {
			return new Exception().getStackTrace()[2].getClassName();
		} catch (Exception e) {
			return "";
		}
	}
	
	public static String localizedString(String str, Object... args)
	{
		String result = localizedStringOrNull(str, args);
		return (result == null ? str : result);
	}
	public static String localizedStringOrNull(String str, Object... args)
	{
		for (ResourceBundle bundle : bundles) {
			try {
				if (args.length < 1)
					return bundle.getString(str);
				return String.format(bundle.getString(str), args);
			} catch (MissingResourceException e) {
			}
		}
		return null;
	}
	
	public static <T extends Comparable<? super T>> T min(T a, T b)
	{
		return (a.compareTo(b) <= 0 ? a : b);
	}
	public static <T extends Comparable<? super T>> T max(T a, T b)
	{
		return (a.compareTo(b) < 0 ? b : a);
	}
	
	public static String capitalize(CharSequence seq)
	{
		String str = seq.toString();
		if (str.length() >= 1 && Character.isLowerCase(str.charAt(0)))
			return Character.toString(Character.toUpperCase(str.charAt(0))) + str.substring(1);
		else
			return str;
	}
	public static String tab()
	{
		return tabString;
	}
	public static String tab(int count)
	{
		StringBuilder result = new StringBuilder();
		while (count-- > 0)
			result.append(tabString);
		return result.toString();
	}
	public static String indent(CharSequence seq)
	{
		return indent(seq, tabString);
	}
	public static String indent(CharSequence seq, CharSequence prefix)
	{
		String str = seq.toString();
		return prefix + str.replaceAll("\n", "\n" + prefix);
	}
	public static String escape(CharSequence seq)
	{
		StringBuilder result = new StringBuilder(seq.length());
		for (int i = 0; i < seq.length(); i++) {
			char chr = seq.charAt(i);
			if (chr == '"')
				result.append("\\\"");
			else if (chr == '\n')
				result.append("\\n");
			else if (chr == '\r')
				result.append("\\r");
			else if (chr == '\t')
				result.append("\\t");
			else if (chr <= 0xFF && !Character.isISOControl(chr))
				result.append(chr);
			else if (chr <= 0xFF)
				result.append("\\").append(String.format("%03o", (int)chr));
			else {
				result.append("\\").append(String.format("%03o", ((int)chr >> 8)))
								   .append(String.format("%03o", ((int)chr & 0xFF)));
			}
		}
		return result.toString();
	}
	public static String escape(byte[] bytes)
	{
		StringBuilder result = new StringBuilder();
		for (byte b : bytes) {
			char c = (char)((char)b & 0xFF);
			if (c == ' ' || ('0' <= c && c <= '9') || ('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z'))
				result.append(c);
			else
				result.append(String.format("\\%03o", (int)c));
		}
		return result.toString();
	}
	
	public static String changeFileSuffix(String file, String from, String to)
	{
		if (file.equals("-"))
			return "-";
		return (file.endsWith(from) ? file.substring(0, file.lastIndexOf(from)) + to : file + to);
	}
	
	public static <T> ArrayList<T> singletonArrayList(T elem)
	{
		ArrayList<T> result = new ArrayList<T>();
		result.add(elem);
		return result;
	}
	
	public static String enumerate(Collection<?> col, CharSequence delimiter)
	{
		StringBuilder result = new StringBuilder();
		int count = 0;
		for (Object obj : col) {
			if (count++ > 0)
				result.append(delimiter);
			result.append(obj);
		}
		return result.toString();
	}
	public static String enumerate(Object[] col, CharSequence delimiter)
	{
		return enumerate(Arrays.asList(col), delimiter);
	}
	
	public static int uniqueID()
	{
		return ++nextID;
	}
	public static String uniqueHash(String seed)
	{
		seed = new Date() + seed + System.currentTimeMillis() + System.nanoTime();
		try {
			StringBuffer result = new StringBuffer();
			String chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_";
			BigInteger n = new BigInteger(1, MessageDigest.getInstance("MD5").digest(seed.getBytes())), 
					   b = BigInteger.valueOf(chars.length());
			while (!n.equals(BigInteger.ZERO)) {
				BigInteger[] x = n.divideAndRemainder(b);
				result.append(chars.charAt(x[1].intValue()));
				n = x[0];
			}
			return result.toString();
		} catch (Exception e) {
			return "_" + System.currentTimeMillis() + System.nanoTime();
		}
	}
}
